"""
(Unofficial) SAD to XSuite Converter

Element Converter
"""

################################################################################
# Required Packages
################################################################################
from tabnanny import verbose
import xtrack as xt
import numpy as np

from scipy.constants import c as clight
from scipy.constants import e as qe

from .._globals import print_section_heading, COORD_SIGNS, TRANSFORM_SHIFT_TOL, TRANSFORM_ROT_TOL, MAX_KNL_ORDER

################################################################################
# RAD2DEG Constant
################################################################################
RAD2DEG = 180.0 / np.pi

################################################################################
# Parsing of strings and floats
################################################################################
def parse_expression(expression: str):
    """
    Try to convert s to float; if that fails, return s stripped
    """
    if type(expression) is float:
        return expression
    elif type(expression) is int:
        return float(expression)
    elif type(expression) is str:
        expression_stripped  = expression.strip()
        try:
            return float(expression_stripped)
        except ValueError:
            return expression_stripped
    else:
        raise TypeError(f"Unsupported type: {type(expression)}. Expected str, int, or float.")

################################################################################
# Check that only one index in knl array is non zero
################################################################################
def only_index_nonzero(
    length: float,
    knl:    list,
    ksl:    list,
    idx:    int,
    tol:    float = 1E-12) -> bool:
    """
    Check that:
      1. length != 0 (within tol)
      2. All entries *except* at index `idx` in both knl and ksl are zero (within tol)
         - Elements may be floats or strings; non-numeric strings count as non-zero.
      3. If require_nonzero_at_idx: at least one of knl[idx], ksl[idx] is non-zero.
    """
    # 1) length check
    if abs(length) <= tol:
        return False

    # helper to test “is this value zero?”
    def is_zero(val) -> bool:
        try:
            return abs(float(val)) <= tol
        except (ValueError, TypeError):
            # non‐numeric ⇒ treat as non‐zero
            return False

    # 2) check every position except idx
    max_len = max(len(knl), len(ksl))
    for arr in (knl, ksl):
        if len(arr) < max_len:
            # pad shorter list with zeros
            arr = arr + [0] * (max_len - len(arr))
        for i, v in enumerate(arr):
            if i == idx:
                continue
            if not is_zero(v):
                return False

    # 3) ensure at least one of knl[idx], ksl[idx] is non‐zero
    if is_zero(knl[idx] if idx < len(knl) else 0) and \
        is_zero(ksl[idx] if idx < len(ksl) else 0):
        return False
    return True

################################################################################
# Get element misalignments
################################################################################
def get_element_misalignments(ele_vars, rotation_correction = 0.0):
        ########################################
        # Define as float zero
        ########################################
        shift_x     = 0.0
        shift_y     = 0.0
        rotation    = 0.0

        ########################################
        # Read values
        ########################################
        if 'dx' in ele_vars:
            shift_x     = parse_expression(ele_vars['dx'])
        if 'dy' in ele_vars:
            shift_y     = parse_expression(ele_vars['dy'])
        if 'rotate' in ele_vars:
            rotation    = parse_expression(ele_vars['rotate'])

        ########################################
        # Rotations in SAD are negative w.r.t. Xsuite
        ########################################
        if isinstance(rotation, str):
            rotation    = f"-{rotation} + {rotation_correction}"
        elif isinstance(rotation, (float, int)):
            rotation    = -rotation + rotation_correction
        else:
            raise TypeError(f"Error reading rotation: type {type(rotation)}")
        

        ########################################
        # Composition of rotations is different in SAD
        ########################################
        if isinstance(rotation, float) and isinstance(shift_x, float) and isinstance(shift_y, float):
            shift_r     = np.sqrt(shift_x**2 + shift_y**2)
            theta_rot   = np.arctan2(shift_y, shift_x)

            shift_x  = shift_r * np.cos(theta_rot - rotation)
            shift_y  = shift_r * np.sin(theta_rot - rotation)
            rotation = rotation
        else:
            shift_x     = str(shift_x)
            shift_y     = str(shift_y)
            rotation    = str(rotation)

            shift_r     = f"sqrt({shift_x}**2 + {shift_y}**2)"
            theta_rot   = f"arctan2({shift_y}, {shift_x})"

            shift_x  = f"{shift_r} * cos({theta_rot} - {rotation})"
            shift_y  = f"{shift_r} * sin({theta_rot} - {rotation})"
            rotation = rotation
        
        return shift_x, shift_y, rotation

################################################################################
# Convert all
################################################################################
def convert_elements(
        parsed_lattice_data:            dict,
        environment:                    xt.Environment,
        user_multipole_replacements:    dict | None = None,
        verbose:                        bool                = False) -> None:
    
    ########################################
    # Get the required data
    ########################################
    parsed_globals  = parsed_lattice_data['globals']
    parsed_elements = parsed_lattice_data['elements']

    ########################################
    # Drifts
    ########################################
    if 'drift' in parsed_elements:
        if verbose:
            print_section_heading("Converting Drifts", mode = 'subsection')
        convert_drifts(parsed_elements, environment)

    ########################################
    # Bends
    ########################################
    if 'bend' in parsed_elements:
        if verbose:
            print_section_heading("Converting Bends", mode = 'subsection')
        convert_bends(parsed_elements, environment)
        convert_correctors(parsed_elements, environment)

    ########################################
    # Quadrupoles
    ########################################
    if 'quad' in parsed_elements:
        if verbose:
            print_section_heading("Converting Quadrupoles", mode = 'subsection')
        convert_quadrupoles(parsed_elements, environment)

    ########################################
    # Sextupoles
    ########################################
    if 'sext' in parsed_elements:
        if verbose:
            print_section_heading("Converting Sextupoles", mode = 'subsection')
        convert_sextupoles(parsed_elements, environment)

    ########################################
    # Octupoles
    ########################################
    if 'oct' in parsed_elements:
        if verbose:
            print_section_heading("Converting Octupoles", mode = 'subsection')
        convert_octupoles(parsed_elements, environment)

    ########################################
    # Multipoles
    ########################################
    if 'mult' in parsed_elements:
        if verbose:
            print_section_heading("Converting Multipoles", mode = 'subsection')
        convert_multipoles(
            parsed_elements             = parsed_elements,
            environment                 = environment,
            user_multipole_replacements = user_multipole_replacements)

    ########################################
    # Cavities
    ########################################
    if 'cavi' in parsed_elements:
        if verbose:
            print_section_heading("Converting Cavities", mode = 'subsection')
        convert_cavities(parsed_elements, environment)

    ########################################
    # Apertures
    ########################################
    if 'apert' in parsed_elements:
        if verbose:
            print_section_heading("Converting Apertures", mode = 'subsection')
        convert_apertures(parsed_elements, environment)

    ########################################
    # Solenoids
    ########################################
    if 'sol' in parsed_elements:
        if verbose:
            print_section_heading("Converting Solenoids", mode = 'subsection')
        convert_solenoids(parsed_elements, parsed_globals, environment)

    ########################################
    # Coordinate Transformations
    ########################################
    if 'coord' in parsed_elements:
        if verbose:
            print_section_heading("Converting Coordinate Transformations", mode = 'subsection')
        convert_coordinate_transformations(parsed_elements, environment)

    ########################################
    # Markers
    ########################################
    if 'mark' in parsed_elements:
        if verbose:
            print_section_heading("Converting Markers", mode = 'subsection')
        convert_markers(parsed_elements, environment)

    ########################################
    # Monitors
    ########################################
    if 'moni' in parsed_elements:
        if verbose:
            print_section_heading("Converting Monitors", mode = 'subsection')
        convert_monitors(parsed_elements, environment)

    ########################################
    # Beam-Beam Interactions
    ########################################
    if 'beambeam' in parsed_elements:
        if verbose:
            print_section_heading("Converting Beam-Beam Interactions", mode = 'subsection')
        convert_beam_beam(parsed_elements, environment)

################################################################################
# Convert drift
################################################################################
def convert_drifts(parsed_elements, environment):
    """
    Convert drifts from the SAD parsed data
    """

    drifts  = parsed_elements['drift']

    for ele_name, ele_vars in drifts.items():

        ########################################
        # Assert Length
        ########################################
        if 'l' in ele_vars:
            length = ele_vars['l']
        else:
            raise ValueError(f"Drift {ele_name} missing length.")

        ########################################
        # Create Element
        ########################################
        environment.new(
            name    = ele_name,
            parent  = xt.Drift,
            length  = length)

################################################################################
# Convert Bends
################################################################################
def convert_bends(parsed_elements, environment):
    """
    Convert bends from the SAD parsed data
    """

    bends  = parsed_elements['bend']

    for ele_name, ele_vars in bends.items():
        if 'angle' in ele_vars:
            
            angle   = parse_expression(ele_vars['angle'])
            if angle == 0:
                continue

            if "l" not in ele_vars:
                # TODO: Improve the handling of this
                k0l             = parse_expression(ele_vars['angle'])
                if k0l != 0:
                    raise ValueError(f"Error! Bend {ele_name} missing length.")
                else:
                    print(f"Warning! Bend {ele_name} missing length and installed as marker")
                    environment.new(
                        name                = ele_name,
                        parent              = xt.Marker)
                    continue

            ########################################
            # Initialise parameters
            ########################################
            length      = 0.0
            k1l         = 0.0
            e1          = 0.0
            e2          = 0.0
            ae1         = 0.0
            ae2         = 0.0

            edge_entry_angle    = 0
            edge_exit_angle     = 0

            ########################################
            # Read values
            ########################################
            length          = float(parse_expression(ele_vars['l']))
            k0l             = parse_expression(ele_vars['angle'])
            
            if 'k1' in ele_vars:
                k1l         = parse_expression(ele_vars['k1'])
            if 'e1' in ele_vars:
                e1          = parse_expression(ele_vars['e1'])
            if 'e2' in ele_vars:
                e2          = parse_expression(ele_vars['e2'])
            if 'ae1' in ele_vars:
                ae1         = parse_expression(ele_vars['ae1'])
            if 'ae2' in ele_vars:
                ae2         = parse_expression(ele_vars['ae2'])

            shift_x, shift_y, rotation  = get_element_misalignments(ele_vars)

            if type(k0l) is float:
                k0  = k0l / length
            else:
                k0  = f"{k0l} / {length}"

            if type(k1l) is float:
                k1  = k1l / length
            else:
                k1  = f"{k1l} / {length}"

            edge_entry_angle    = f"{e1} * {k0l} + {ae1}"
            edge_exit_angle     = f"{e2} * {k0l} + {ae2}"

            ########################################
            # Create variables
            ########################################
            environment[f'k0_{ele_name}']   = k0
            k0                              = f"k0_{ele_name}"
            
            if k1 != 0:
                environment[f'k1_{ele_name}']   = k1
                k1                              = f"k1_{ele_name}"

            ########################################
            # Create Element
            ########################################
            environment.new(
                name                = ele_name,
                parent              = xt.Bend,
                length              = length,
                k0                  = k0,
                k1                  = k1,
                h                   = k0,
                edge_entry_angle    = edge_entry_angle,
                edge_exit_angle     = edge_exit_angle,
                shift_x             = shift_x,
                shift_y             = shift_y,
                rot_s_rad           = rotation)
            continue

################################################################################
# Convert Correctors
################################################################################
def convert_correctors(parsed_elements, environment):
    """
    Convert correctors from the SAD parsed data
    """

    bends  = parsed_elements['bend']

    for ele_name, ele_vars in bends.items():

        is_corrector    = False
        if 'angle' in ele_vars:
            angle   = parse_expression(ele_vars['angle'])
            if angle == 0:
                is_corrector    = True
        if 'angle' not in ele_vars:
            is_corrector    = True

        if is_corrector:

            ########################################
            # Initialise parameters
            ########################################
            length      = 0.0
            k0l         = 0.0

            ########################################
            # Read values
            ########################################
            if 'l' in ele_vars:
                length      = parse_expression(ele_vars['l'])

            shift_x, shift_y, rotation  = get_element_misalignments(ele_vars)

            if length == 0:
                print(f"Warning! Corrector {ele_name} missing length and installed as marker")

                environment.new(
                    name    = ele_name,
                    parent  = xt.Marker)
                continue

            if 'k0' in ele_vars:
                k0l             = parse_expression(ele_vars['k0'])
            if type(k0l) is float:
                k0  = k0l / ele_vars['l']
            else:
                k0  = f"{k0l} / {ele_vars['l']}"

            ########################################
            # Create variables
            ########################################
            environment[f'k0_{ele_name}']   = k0
            k0                              = f"k0_{ele_name}"

            ########################################
            # Create Element
            ########################################
            environment.new(
                name                = ele_name,
                parent              = xt.Bend,
                length              = length,
                k0                  = k0,
                k1                  = 0.0,
                h                   = 0.0,
                edge_entry_angle    = 0.0,
                edge_exit_angle     = 0.0,
                shift_x             = shift_x,
                shift_y             = shift_y,
                rot_s_rad           = rotation)
            continue

################################################################################
# Convert Quadrupoles
################################################################################
def convert_quadrupoles(parsed_elements, environment):
    """
    Convert quadrupoles from the SAD parsed data
    """

    quads  = parsed_elements['quad']

    for ele_name, ele_vars in quads.items():

        ########################################
        # Initialise parameters
        ########################################
        length      = 0.0
        k1l         = 0.0
        k1sl        = 0.0

        ########################################
        # Read values
        ########################################
        if 'l' in ele_vars:
            length      = parse_expression(ele_vars['l'])
        else:
            raise ValueError(f"Error! Quadrupole {ele_name} missing length.")

        shift_x, shift_y, rotation  = get_element_misalignments(ele_vars)

        if 'k1' in ele_vars:
            if not isinstance(rotation, float):
                k1l     = f"{ele_vars['k1']}"
            else:

                if np.isclose(rotation, +np.pi / 4, atol = 1E-6):
                    if isinstance(ele_vars['k1'], (float, int)):
                        k1sl    = -ele_vars['k1']
                    else:
                        k1sl    = f"-{ele_vars['k1']}"
                    shift_x, shift_y, rotation  = get_element_misalignments(
                        ele_vars            = ele_vars,
                        rotation_correction = -np.pi / 4)

                elif np.isclose(rotation, -np.pi / 4, atol = 1E-6):
                    if isinstance(ele_vars['k1'], (float, int)):
                        k1sl    = +ele_vars['k1']
                    else:
                        k1sl    = f"+{ele_vars['k1']}"
                    shift_x, shift_y, rotation  = get_element_misalignments(
                        ele_vars            = ele_vars,
                        rotation_correction = +np.pi / 4)

                else:
                    k1l     = ele_vars['k1']

        if isinstance(k1l, float):
            k1  = k1l / ele_vars['l']
        else:
            k1  = f"{k1l} / {ele_vars['l']}"
        
        if isinstance(k1sl, float):
            k1s = k1sl / ele_vars['l']
        else:
            k1s = f"{k1sl} / {ele_vars['l']}"

        ########################################
        # Create variables
        ########################################
        if k1 != 0:
            environment[f'k1_{ele_name}']   = k1
            k1                              = f"k1_{ele_name}"
        if k1s != 0:
            environment[f'k1s_{ele_name}']  = k1s
            k1s                             = f"k1s_{ele_name}"

        ########################################
        # Create Element
        ########################################
        environment.new(
            name        = ele_name,
            parent      = xt.Quadrupole,
            length      = length,
            k1          = k1,
            k1s         = k1s,
            shift_x     = shift_x,
            shift_y     = shift_y,
            rot_s_rad   = rotation)
        continue

################################################################################
# Convert Sextupoles
################################################################################
def convert_sextupoles(parsed_elements, environment):
    """
    Convert sextupoles from the SAD parsed data
    """

    sexts  = parsed_elements['sext']

    for ele_name, ele_vars in sexts.items():

        ########################################
        # Initialise parameters
        ########################################
        length      = 0.0
        k2l         = 0.0
        k2sl        = 0.0

        ########################################
        # Read values
        ########################################
        if 'l' in ele_vars:
            length      = parse_expression(ele_vars['l'])
        else:
            raise ValueError(f"Error! Sextupole {ele_name} missing length.")

        shift_x, shift_y, rotation  = get_element_misalignments(ele_vars)

        if 'k2' in ele_vars:
            if not isinstance(rotation, float):
                k2l     = f"{ele_vars['k2']}"
            else:

                if np.isclose(rotation, +np.pi / 6, atol = 1E-6):
                    if isinstance(ele_vars['k2'], (float, int)):
                        k2sl    = -ele_vars['k2']
                    else:
                        k2sl    = f"-{ele_vars['k2']}"
                    shift_x, shift_y, rotation  = get_element_misalignments(
                        ele_vars            = ele_vars,
                        rotation_correction = -np.pi / 6)

                elif np.isclose(rotation, -np.pi / 6, atol = 1E-6):
                    if isinstance(ele_vars['k2'], (float, int)):
                        k2sl    = +ele_vars['k2']
                    else:
                        k2sl    = f"+{ele_vars['k2']}"
                    shift_x, shift_y, rotation  = get_element_misalignments(
                        ele_vars            = ele_vars,
                        rotation_correction = +np.pi / 6)

                else:
                    k2l     = ele_vars['k2']

        if isinstance(k2l, float):
            k2  = k2l / ele_vars['l']
        else:
            k2  = f"{k2l} / {ele_vars['l']}"

        if isinstance(k2sl, float):
            k2s = k2sl / ele_vars['l']
        else:
            k2s = f"{k2sl} / {ele_vars['l']}"

        ########################################
        # Create variables
        ########################################
        if k2 != 0:
            environment[f'k2_{ele_name}']   = k2
            k2                              = f"k2_{ele_name}"
        if k2s != 0:
            environment[f'k2s_{ele_name}']  = k2s
            k2s                             = f"k2s_{ele_name}"

        ########################################
        # Create Element
        ########################################
        environment.new(
            name        = ele_name,
            parent      = xt.Sextupole,
            length      = length,
            k2          = k2,
            k2s         = k2s,
            shift_x     = shift_x,
            shift_y     = shift_y,
            rot_s_rad   = rotation)
        continue

################################################################################
# Convert Octupoles
################################################################################
def convert_octupoles(parsed_elements, environment):
    """
    Convert octupoles from the SAD parsed data
    """

    octs    = parsed_elements['oct']

    for ele_name, ele_vars in octs.items():

        if "l" not in ele_vars:
            # TODO: Improve the handling of this
            print(f"Warning! Octupole {ele_name} missing length and installed as marker")
            environment.new(
                name                = ele_name,
                parent              = xt.Marker)
            continue

        ########################################
        # Initialise parameters
        ########################################
        length      = 0.0
        k3l         = 0.0
        k3sl        = 0.0

        ########################################
        # Read values
        ########################################
        if 'l' in ele_vars:
            length      = parse_expression(ele_vars['l'])
        else:
            raise ValueError(f"Error! Octupole {ele_name} missing length.")

        shift_x, shift_y, rotation  = get_element_misalignments(ele_vars)

        if 'k3' in ele_vars:
            if not isinstance(rotation, float):
                k3l     = f"{ele_vars['k3']}"
            else:

                if np.isclose(rotation, +np.pi / 8, atol = 1E-6):
                    if isinstance(ele_vars['k3'], (float, int)):
                        k3sl    = -ele_vars['k3']
                    else:
                        k3sl    = f"-{ele_vars['k3']}"
                    shift_x, shift_y, rotation  = get_element_misalignments(
                        ele_vars            = ele_vars,
                        rotation_correction = -np.pi / 8)

                elif np.isclose(rotation, -np.pi / 8, atol = 1E-6):
                    if isinstance(ele_vars['k3'], (float, int)):
                        k3sl    = +ele_vars['k3']
                    else:
                        k3sl    = f"+{ele_vars['k3']}"
                    shift_x, shift_y, rotation  = get_element_misalignments(
                        ele_vars            = ele_vars,
                        rotation_correction = +np.pi / 8)

                else:
                    k3l     = ele_vars['k3']

        if isinstance(k3l, float):
            k3  = k3l / ele_vars['l']
        else:
            k3  = f"{k3l} / {ele_vars['l']}"

        if isinstance(k3sl, float):
            k3s = k3sl / ele_vars['l']
        else:
            k3s = f"{k3sl} / {ele_vars['l']}"

        ########################################
        # Create variables
        ########################################
        if k3 != 0:
            environment[f'k3_{ele_name}']   = k3
            k3                              = f"k3_{ele_name}"
        if k3s != 0:
            environment[f'k3s_{ele_name}']  = k3s
            k3s                             = f"k3s_{ele_name}"

        ########################################
        # Create Element
        ########################################
        environment.new(
            name        = ele_name,
            parent      = xt.Octupole,
            length      = length,
            k3          = k3,
            k3s         = k3s,
            shift_x     = shift_x,
            shift_y     = shift_y,
            rot_s_rad   = rotation)
        continue

################################################################################
# Convert Multipoles
################################################################################
def convert_multipoles(
        parsed_elements,
        environment,
        user_multipole_replacements = None,
        simplify_multipoles         = True):
    """
    Convert multipoles from the SAD parsed data
    """

    mults   = parsed_elements['mult']

    for ele_name, ele_vars in mults.items():

        ########################################
        # Initialise parameters
        ########################################
        length      = 0.0

        ########################################
        # Read values
        ########################################
        if 'l' in ele_vars:
            length      = parse_expression(ele_vars['l'])

        shift_x, shift_y, rotation  = get_element_misalignments(ele_vars)

        knl = []
        for kn in range(0, MAX_KNL_ORDER):
            knl.append(0.0)
            if f'k{kn}' in ele_vars:
                knl[kn] = parse_expression(ele_vars[f'k{kn}'])

        ksl = []
        for ks in range(0, MAX_KNL_ORDER):
            ksl.append(0.0)
            if f'sk{ks}' in ele_vars:
                ksl[ks] = parse_expression(ele_vars[f'sk{ks}'])

        ########################################
        # User Defined Multipole Replacements
        ########################################
        if user_multipole_replacements is not None:
            if any(ele_name.startswith(test_key) for test_key in user_multipole_replacements):
                replace_type    = None

                if not 'l' in ele_vars:
                    print(f'Warning! Multipole {ele_name} is a thin lens, replacement not supported for thin lens')
                    continue

                # Search the multipole replacements dict for the type of element
                for replacement in user_multipole_replacements:
                    if ele_name.startswith(replacement):
                        replace_type    = user_multipole_replacements[replacement]

                ########################################
                # Bend Replacement (kick)
                ########################################
                if replace_type == 'Bend':

                    if knl[0] != 0 and ksl[0] != 0:
                        if type(knl[0]) is float or type(ksl[0]) is float:
                            k0l         = f"sqrt({knl[0]}**2 + {ksl[0]}**2)"
                            rotation    = f"{rotation} + arctan2({ksl[0]}, {knl[0]})"
                        else:
                            k0l         = np.sqrt(knl[0]**2 + ksl[0]**2)
                            rotation    = rotation + np.arctan2(ksl[0], knl[0])
                    elif knl[0] != 0:
                        k0l         = knl[0]
                        rotation    = rotation
                    elif ksl[0] != 0:
                        k0l         = ksl[0]
                        if type(rotation) is float:
                            rotation    = rotation + np.pi / 2
                        else:
                            rotation    = f"{rotation} + np.pi / 2"
                    else:
                        k0l = 0.0
                    
                    if type(k0l) is float:
                        k0  = k0l / ele_vars['l']
                    else:
                        k0  = f"{k0l} / {ele_vars['l']}"

                    ####################
                    # Create variables
                    ####################
                    if k0 != 0:
                        environment[f'k0_{ele_name}']   = k0
                        k0                              = f"k0_{ele_name}"

                    ####################
                    # Create Element
                    ####################
                    environment.new(
                        name                = ele_name,
                        parent              = xt.Bend,
                        length              = length,
                        k0                  = k0,
                        shift_x             = shift_x,
                        shift_y             = shift_y,
                        rot_s_rad           = rotation)
                    continue

                ########################################
                # Quadrupole Replacement
                ########################################
                elif replace_type == 'Quadrupole':

                    k1l     = knl[1]
                    k1sl    = ksl[1]

                    if type(k1l) is float:
                        k1  = k1l / ele_vars['l']
                    else:
                        k1  = f"{k1l} / {ele_vars['l']}"
                    if type(k1sl) is float:
                        k1s = k1sl / ele_vars['l']
                    else:
                        k1s = f"{k1sl} / {ele_vars['l']}"

                    ####################
                    # Create variables
                    ####################
                    if k1 != 0:
                        environment[f'k1_{ele_name}']   = k1
                        k1                              = f"k1_{ele_name}"
                    if k1s != 0:
                        environment[f'k1s_{ele_name}']  = k1s
                        k1s                             = f"k1s_{ele_name}"

                    ####################
                    # Create Element
                    ####################
                    environment.new(
                        name                = ele_name,
                        parent              = xt.Quadrupole,
                        length              = length,
                        k1                  = k1,
                        k1s                 = k1s,
                        shift_x             = shift_x,
                        shift_y             = shift_y,
                        rot_s_rad           = rotation)
                    continue

                ########################################
                # Sextupole Replacement
                ########################################
                elif replace_type == 'Sextupole':

                    k2l     = knl[2]
                    k2sl    = ksl[2]

                    if type(k2l) is float:
                        k2  = k2l / ele_vars['l']
                    else:
                        k2  = f"{k2l} / {ele_vars['l']}"
                    if type(k2sl) is float:
                        k2s = k2sl / ele_vars['l']
                    else:
                        k2s = f"{k2sl} / {ele_vars['l']}"

                    ####################
                    # Create variables
                    ####################
                    if k2 != 0:
                        environment[f'k2_{ele_name}']   = k2
                        k2                              = f"k2_{ele_name}"
                    if k2s != 0:
                        environment[f'k2s_{ele_name}']  = k2s
                        k2s                             = f"k2s_{ele_name}"

                    ####################
                    # Create Element
                    ####################
                    environment.new(
                        name                = ele_name,
                        parent              = xt.Sextupole,
                        length              = length,
                        k2                  = k2,
                        k2s                 = k2s,
                        shift_x             = shift_x,
                        shift_y             = shift_y,
                        rot_s_rad           = rotation)
                    continue

                ########################################
                # Octupole Replacement
                ########################################
                elif replace_type == 'Octupole':

                    k3l     = knl[3]
                    k3sl    = ksl[3]

                    if type(k3l) is float:
                        k3  = k3l / ele_vars['l']
                    else:
                        k3  = f"{k3l} / {ele_vars['l']}"
                    if type(k3sl) is float:
                        k3s = k3sl / ele_vars['l']
                    else:
                        k3s = f"{k3sl} / {ele_vars['l']}"

                    ####################
                    # Create variables
                    ####################
                    if k3 != 0:
                        environment[f'k3_{ele_name}']   = k3
                        k3                              = f"k3_{ele_name}"
                    if k3s != 0:
                        environment[f'k3s_{ele_name}']  = k3s
                        k3s                             = f"k3s_{ele_name}"

                    ####################
                    # Create Element
                    ####################
                    environment.new(
                        name                = ele_name,
                        parent              = xt.Octupole,
                        length              = length,
                        k3                  = k3,
                        k3s                 = k3s,
                        shift_x             = shift_x,
                        shift_y             = shift_y,
                        rot_s_rad           = rotation)
                    continue
                else:
                    raise ValueError('Error: Unknown element replacement')

        ########################################
        # Automatic Simplification
        ########################################
        if simplify_multipoles:

            ########################################
            # Correctors stored as multipoles
            ########################################
            if only_index_nonzero(
                    length  = float(length),
                    knl     = knl,
                    ksl     = ksl,
                    idx     = 0):
                
                if knl[0] != 0 and ksl[0] != 0:
                    if type(knl[0]) is float or type(ksl[0]) is float:
                        k0l         = f"sqrt({knl[0]}**2 + {ksl[0]}**2)"
                        rotation    = f"{rotation} + arctan2({ksl[0]}, {knl[0]})"
                    else:
                        k0l         = np.sqrt(knl[0]**2 + ksl[0]**2)
                        rotation    = rotation + np.arctan2(ksl[0], knl[0])
                elif knl[0] != 0:
                    k0l         = knl[0]
                    rotation    = rotation
                elif ksl[0] != 0:
                    k0l         = ksl[0]
                    if type(rotation) is float:
                        rotation    = rotation + np.pi / 2
                    else:
                        rotation    = f"{rotation} + np.pi / 2"
                else:
                    k0l = 0
                
                if type(k0l) is float:
                    k0  = k0l / ele_vars['l']
                else:
                    k0  = f"{k0l} / {ele_vars['l']}"

                ####################
                # Create variables
                ####################
                if k0 != 0:
                    environment[f'k0_{ele_name}']   = k0
                    k0                              = f"k0_{ele_name}"

                ####################
                # Create Element
                ####################
                environment.new(
                    name                = ele_name,
                    parent              = xt.Bend,
                    length              = length,
                    k0                  = k0,
                    shift_x             = shift_x,
                    shift_y             = shift_y,
                    rot_s_rad           = rotation)
                continue

            ########################################
            # Quadrupoles stored as multipoles
            ########################################
            if only_index_nonzero(
                    length  = float(length),
                    knl     = knl,
                    ksl     = ksl,
                    idx     = 1):

                k1l     = knl[1]
                k1sl    = ksl[1]

                if type(k1l) is float:
                    k1  = k1l / ele_vars['l']
                else:
                    k1  = f"{k1l} / {ele_vars['l']}"
                if type(k1sl) is float:
                    k1s = k1sl / ele_vars['l']
                else:
                    k1s = f"{k1sl} / {ele_vars['l']}"

                ####################
                # Create variables
                ####################
                if k1 != 0:
                    environment[f'k1_{ele_name}']   = k1
                    k1                              = f"k1_{ele_name}"
                if k1s != 0:
                    environment[f'k1s_{ele_name}']  = k1s
                    k1s                             = f"k1s_{ele_name}"

                ####################
                # Create Element
                ####################
                environment.new(
                    name                = ele_name,
                    parent              = xt.Quadrupole,
                    length              = length,
                    k1                  = k1,
                    k1s                 = k1s,
                    shift_x             = shift_x,
                    shift_y             = shift_y,
                    rot_s_rad           = rotation)
                continue

            ########################################
            # Sextupoles stored as multipoles
            ########################################
            if only_index_nonzero(
                    length  = float(length),
                    knl     = knl,
                    ksl     = ksl,
                    idx     = 2):

                k2l     = knl[2]
                k2sl    = ksl[2]

                if type(k2l) is float:
                    k2  = k2l / ele_vars['l']
                else:
                    k2  = f"{k2l} / {ele_vars['l']}"
                if type(k2sl) is float:
                    k2s = k2sl / ele_vars['l']
                else:
                    k2s = f"{k2sl} / {ele_vars['l']}"

                ####################
                # Create variables
                ####################
                if k2 != 0:
                    environment[f'k2_{ele_name}']   = k2
                    k2                              = f"k2_{ele_name}"
                if k2s != 0:
                    environment[f'k2s_{ele_name}']  = k2s
                    k2s                             = f"k2s_{ele_name}"

                ####################
                # Create Element
                ####################
                environment.new(
                    name                = ele_name,
                    parent              = xt.Sextupole,
                    length              = length,
                    k2                  = k2,
                    k2s                 = k2s,
                    shift_x             = shift_x,
                    shift_y             = shift_y,
                    rot_s_rad           = rotation)
                continue

            ########################################
            # Octupoles stored as multipoles
            ########################################
            if only_index_nonzero(
                    length  = float(length),
                    knl     = knl,
                    ksl     = ksl,
                    idx     = 3):

                k3l     = knl[3]
                k3sl    = ksl[3]

                if type(k3l) is float:
                    k3  = k3l / ele_vars['l']
                else:
                    k3  = f"{k3l} / {ele_vars['l']}"
                if type(k3sl) is float:
                    k3s = k3sl / ele_vars['l']
                else:
                    k3s = f"{k3sl} / {ele_vars['l']}"

                ####################
                # Create variables
                ####################
                if k3 != 0:
                    environment[f'k3_{ele_name}']   = k3
                    k3                              = f"k3_{ele_name}"
                if k3s != 0:
                    environment[f'k3s_{ele_name}']  = k3s
                    k3s                             = f"k3s_{ele_name}"

                ####################
                # Create Element
                ####################
                environment.new(
                    name                = ele_name,
                    parent              = xt.Octupole,
                    length              = length,
                    k3                  = k3,
                    k3s                 = k3s,
                    shift_x             = shift_x,
                    shift_y             = shift_y,
                    rot_s_rad           = rotation)
                continue

        ########################################
        # True multipole element
        ########################################
        environment.new(
            name        = ele_name,
            parent      = xt.Multipole,
            _isthick    = True,
            length      = length,
            knl         = knl,
            ksl         = ksl,
            order       = MAX_KNL_ORDER,
            shift_x     = shift_x,
            shift_y     = shift_y,
            rot_s_rad   = rotation)
        continue

################################################################################
# Convert Cavities
################################################################################
def convert_cavities(parsed_elements, environment):
    """
    Convert cavities from the SAD parsed data
    """

    cavis   = parsed_elements['cavi']

    for ele_name, ele_vars in cavis.items():

        ########################################
        # Initialise parameters
        ########################################
        length      = 0.0
        voltage     = 0.0
        freq        = 0.0
        phi         = 180.0

        ########################################
        # Read values
        ########################################
        if 'l' in ele_vars:
            length      = parse_expression(ele_vars['l'])
        if 'volt' in ele_vars:
            voltage = parse_expression(ele_vars['volt'])
        if 'freq' in ele_vars:
            freq = parse_expression(ele_vars['freq'])
        if 'phi' in ele_vars:
            phi_offset = parse_expression(ele_vars['phi'])
            if type(phi_offset) is float:
                phi_offset  = np.rad2deg(phi_offset)
                phi         += phi_offset
            elif type(phi_offset) is str:
                phi_offset  = f"({RAD2DEG} * {phi_offset})"
                phi         = f"{phi} + {phi_offset}"
            else:
                raise ValueError(f"Unsupported type for phi offset: {type(phi_offset)}")

        if 'harm' in ele_vars:
            print(f"Cavity {ele_name} is harmonic and addressed later")

        ########################################
        # Create variables
        ########################################
        environment[f'vol_{ele_name}']      = voltage
        
        if freq != 0:
            environment[f'freq_{ele_name}'] = freq
            freq                            = f"freq_{ele_name} * (1 + fshift)"
        if phi != 0:
            environment[f'lag_{ele_name}']  = phi
            phi                             = f"lag_{ele_name}"

        ########################################
        # Create Element
        ########################################
        environment.new(
            name        = ele_name,
            parent      = xt.Cavity,
            length      = length,
            voltage     = voltage,
            frequency   = freq,
            lag         = phi)
        continue

################################################################################
# Convert Apertures
################################################################################
def convert_apertures(parsed_elements, environment):
    """
    Convert apertures from the SAD parsed data
    """

    aperts  = parsed_elements['apert']

    for ele_name, ele_vars in aperts.items():

        ########################################
        # Initialise parameters
        ########################################
        a       = 1.0
        b       = 1.0

        ########################################
        # Read values
        ########################################
        if 'ax' in ele_vars:
            a = parse_expression(ele_vars['ax'])
        if 'ay' in ele_vars:
            b = parse_expression(ele_vars['ay'])

        ########################################
        # Create Element
        ########################################
        environment.new(
            name    = ele_name,
            parent  = xt.LimitEllipse,
            a       = a,
            b       = b)
        continue

################################################################################
# Convert Solenoids
################################################################################
def convert_solenoids(parsed_elements, parsed_globals, environment):
    """
    Convert solenoids from the SAD parsed data
    """

    P0_J    = parsed_globals['momentum'] * qe / clight
    BRHO    = P0_J / qe / parsed_globals['charge']

    solenoids   = parsed_elements['sol']

    for ele_name, ele_vars in solenoids.items():

        ########################################
        # Initialise parameters
        ########################################
        bound       = False
        geo         = False

        offset_x    = 0.0
        offset_y    = 0.0
        offset_z    = 0.0
        rot_chi1    = 0.0
        rot_chi2    = 0.0
        rot_chi3    = 0.0

        # Per Oide, there is no offset s

        ########################################
        # Read values
        ########################################
        bz  = parse_expression(ele_vars['bz'])
        ks  = bz / BRHO

        if 'bound' in ele_vars:
            bound   = True
        else:
            bound   = False

        if 'geo' in ele_vars:
            geo     = True
        else:
            geo     = False

        # Based on testing, when geo, use the dpx, dpy etc
        if 'dx' in ele_vars:
            offset_x    = parse_expression(ele_vars['dx'])
        if 'dy' in ele_vars:
            offset_y    = parse_expression(ele_vars['dy'])
        if 'dz' in ele_vars:
            offset_z    = parse_expression(ele_vars['dz'])
            offset_s    = parse_expression(ele_vars['dz'])
        if 'dpx' in ele_vars:
            rot_chi1    = parse_expression(ele_vars['dpx'])
        if 'dpy' in ele_vars:
            rot_chi2    = parse_expression(ele_vars['dpy'])

        if not geo:
            # Then use the other rotations
            if ('dpx' not in ele_vars) and ('chi1' in ele_vars):
                rot_chi1    = parse_expression(ele_vars['chi1'])
            if ('dpy' not in ele_vars) and ('chi2' in ele_vars):
                rot_chi2    = parse_expression(ele_vars['chi2'])
            if ('dpz' not in ele_vars) and ('chi3' in ele_vars):
                rot_chi3    = parse_expression(ele_vars['chi3'])

        # Should not have dz in geo sol
        if geo and 'dz' in ele_vars:
            if verbose:
                print(f"Warning! Solenoid {ele_name} is a geo solenoid but with dz defined: ignoring dz")
            offset_z = 0.0

        ########################################
        # Zero small values
        ########################################
        if isinstance(offset_x, float) and np.abs(offset_x) < TRANSFORM_SHIFT_TOL:
            offset_x = 0.0
        if isinstance(offset_y, float) and np.abs(offset_y) < TRANSFORM_SHIFT_TOL:
            offset_y = 0.0
        if isinstance(offset_z, float) and np.abs(offset_z) < TRANSFORM_SHIFT_TOL:
            offset_z = 0.0
        if isinstance(rot_chi1, float) and np.abs(rot_chi1) < TRANSFORM_ROT_TOL:
            rot_chi1 = 0.0
        if isinstance(rot_chi2, float) and np.abs(rot_chi2) < TRANSFORM_ROT_TOL:
            rot_chi2 = 0.0
        if isinstance(rot_chi3, float) and np.abs(rot_chi3) < TRANSFORM_ROT_TOL:
            rot_chi3 = 0.0

        ########################################
        # Shift Transforms
        ########################################
        SOL_DX_FACTOR   = -1 * COORD_SIGNS['dx']
        SOL_DY_FACTOR   = -1 * COORD_SIGNS['dy']
        SOL_DZ_FACTOR   = -1

        if type(offset_x) is float:
            offset_x    = SOL_DX_FACTOR * offset_x
        elif type(offset_x) is str:
            offset_x    = f"{SOL_DX_FACTOR} * {offset_x}"
        else:
            raise ValueError(f"Unsupported type for offset_x: {type(offset_x)}")
        
        if type(offset_y) is float:
            offset_y    = SOL_DY_FACTOR * offset_y
        elif type(offset_y) is str:
            offset_y    = f"{SOL_DY_FACTOR} * {offset_y}"
        else:
            raise ValueError(f"Unsupported type for offset_y: {type(offset_y)}")

        if type(offset_z) is float:
            offset_z    = SOL_DZ_FACTOR * offset_z
        elif type(offset_z) is str:
            offset_z    = f"{SOL_DZ_FACTOR} * {offset_z}"
        else:
            raise ValueError(f"Unsupported type for offset_z: {type(offset_z)}")

        ########################################
        # Angle Transforms
        ########################################
        SOL_CHI1_FACTOR = -1 * COORD_SIGNS['chi1']
        SOL_CHI2_FACTOR = -1 * COORD_SIGNS['chi2']
        SOL_CHI3_FACTOR = -1 * COORD_SIGNS['chi3']

        if type(rot_chi1) is float:
            rot_chi1    = np.rad2deg(SOL_CHI1_FACTOR * rot_chi1)
        elif type(rot_chi1) is str:
            rot_chi1    = f"{SOL_CHI1_FACTOR} * {rot_chi1} * {RAD2DEG}"
        else:
            raise ValueError(f"Unsupported type for rot_chi1: {type(rot_chi1)}")
        
        if type(rot_chi2) is float:
            rot_chi2    = np.rad2deg(SOL_CHI2_FACTOR * rot_chi2)
        elif type(rot_chi2) is str:
            rot_chi2    = f"{SOL_CHI2_FACTOR} * {rot_chi2} * {RAD2DEG}"
        else:
            raise ValueError(f"Unsupported type for rot_chi2: {type(rot_chi2)}")
        
        if type(rot_chi3) is float:
            rot_chi3    = np.rad2deg(SOL_CHI3_FACTOR * rot_chi3)
        elif type(rot_chi3) is str:
            rot_chi3    = f"{SOL_CHI3_FACTOR} * {rot_chi3} * {RAD2DEG}"
        else:
            raise ValueError(f"Unsupported type for rot_chi3: {type(rot_chi3)}")

        ########################################
        # Compound Solenoid Element
        ########################################
        if bound:

            ########################################
            # Create the elements
            ########################################
            environment.new(
                name    = f'{ele_name}_bound',
                parent  = xt.Solenoid,
                ks      = ks)

            environment.new(
                name    = f'{ele_name}_dxy',
                parent  = xt.XYShift,
                dx      = offset_x,
                dy      = offset_y)
            
            environment.new(
                name    = f'{ele_name}_dz',
                parent  = xt.ZetaShift,
                dzeta   = offset_z)

            environment.new(
                name    = f'{ele_name}_chi2',
                parent  = xt.XRotation,
                angle   = rot_chi2)

            environment.new(
                name    = f'{ele_name}_chi1',
                parent  = xt.YRotation,
                angle   = rot_chi1)

            environment.new(
                name    = f'{ele_name}_chi3',
                parent  = xt.SRotation,
                angle   = rot_chi3)

            # ds kills things w.r.t sad
            # Considering the survey, it breaks everything
            # There is a discrepancy in s when it is removed
            # This is just because SAD takes dz into account with s
            # SAD s is based on particle path!

            ########################################
            # Order the elements (reordered later)
            ########################################
            compound_solenoid_components = [
                f'{ele_name}_bound',
                f'{ele_name}_dxy',
                f'{ele_name}_dz',
                f'{ele_name}_chi1',
                f'{ele_name}_chi2',
                f'{ele_name}_chi3']
            environment.new_line(
                name        = ele_name,
                components  = compound_solenoid_components)
            continue
        else:
            environment.new(
                name    = f'{ele_name}',
                parent  = xt.Solenoid,
                ks      = ks)
            continue

################################################################################
# Convert Markers
################################################################################
def convert_markers(parsed_elements, environment):
    """
    Convert markers from the SAD parsed data
    """

    markers   = parsed_elements['mark']

    for ele_name, ele_vars in markers.items():

        ########################################
        # Create Element
        ########################################
        environment.new(
                name    = ele_name,
                parent  = xt.Marker)
        continue

################################################################################
# Convert Monitors
################################################################################
def convert_monitors(parsed_elements, environment):
    """
    Convert monitors from the SAD parsed data
    """

    monitors   = parsed_elements['moni']

    for ele_name, ele_vars in monitors.items():

        ########################################
        # Create Element
        ########################################
        environment.new(
                name    = ele_name,
                parent  = xt.Marker)
        continue

################################################################################
# Convert Beam-Beam Interactions
################################################################################
def convert_beam_beam(parsed_elements, environment):
    """
    Convert beam-beam interactions from the SAD parsed data
    """

    beam_beams   = parsed_elements['beambeam']

    for ele_name, ele_vars in beam_beams.items():

        ########################################
        # Create Element
        ########################################
        environment.new(
                name    = ele_name,
                parent  = xt.Marker)
        continue

################################################################################
# Convert Coordinate Transformations
################################################################################
def convert_coordinate_transformations(parsed_elements, environment):
    """
    Convert coordinate transformations from the SAD parsed data
    """

    coord_transforms   = parsed_elements['coord']
    for ele_name, ele_vars in coord_transforms.items():

        ########################################
        # Initialise parameters
        ########################################
        n_transforms    = 0

        offset_x    = 0.0
        offset_y    = 0.0
        rot_chi1    = 0.0
        rot_chi2    = 0.0
        rot_chi3    = 0.0

        ########################################
        # Read values
        ########################################
        if 'dx' in ele_vars:
            offset_x    = parse_expression(ele_vars['dx'])
        if 'dy' in ele_vars:
            offset_y    = parse_expression(ele_vars['dy'])
        if 'chi1' in ele_vars:
            rot_chi1    = parse_expression(ele_vars['chi1'])
        if 'chi2' in ele_vars:
            rot_chi2    = parse_expression(ele_vars['chi2'])
        if 'chi3' in ele_vars:
            rot_chi3    = parse_expression(ele_vars['chi3'])

        ########################################
        # Zero small values
        ########################################
        if isinstance(offset_x, float) and np.abs(offset_x) < TRANSFORM_SHIFT_TOL:
            offset_x = 0.0
        if isinstance(offset_y, float) and np.abs(offset_y) < TRANSFORM_SHIFT_TOL:
            offset_y = 0.0
        if isinstance(rot_chi1, float) and np.abs(rot_chi1) < TRANSFORM_ROT_TOL:
            rot_chi1 = 0.0
        if isinstance(rot_chi2, float) and np.abs(rot_chi2) < TRANSFORM_ROT_TOL:
            rot_chi2 = 0.0
        if isinstance(rot_chi3, float) and np.abs(rot_chi3) < TRANSFORM_ROT_TOL:
            rot_chi3 = 0.0

        ########################################
        # Count Transforms
        ########################################
        if offset_x != 0:
            n_transforms += 1
        if offset_y != 0:
            n_transforms += 1
        if rot_chi1 != 0:
            n_transforms += 1
        if rot_chi2 != 0:
            n_transforms += 1
        if rot_chi3 != 0:
            n_transforms += 1

        ########################################
        # Shift Transforms
        ########################################
        COORD_DX_FACTOR   = COORD_SIGNS['dx']
        COORD_DY_FACTOR   = COORD_SIGNS['dy']

        if type(offset_x) is float:
            offset_x    = COORD_DX_FACTOR * offset_x
        elif type(offset_x) is str:
            offset_x    = f"{COORD_DX_FACTOR} * {offset_x}"
        else:
            raise ValueError(f"Unsupported type for offset_x: {type(offset_x)}")
        
        if type(offset_y) is float:
            offset_y    = COORD_DY_FACTOR * offset_y
        elif type(offset_y) is str:
            offset_y    = f"{COORD_DY_FACTOR} * {offset_y}"
        else:
            raise ValueError(f"Unsupported type for offset_y: {type(offset_y)}")

        ########################################
        # Angle Transforms
        ########################################
        COORD_CHI1_FACTOR   = COORD_SIGNS['chi1']
        COORD_CHI2_FACTOR   = COORD_SIGNS['chi2']
        COORD_CHI3_FACTOR   = COORD_SIGNS['chi3']

        if type(rot_chi1) is float:
            rot_chi1    = np.rad2deg(COORD_CHI1_FACTOR * rot_chi1)
        elif type(rot_chi1) is str:
            rot_chi1    = f"{COORD_CHI1_FACTOR} * {rot_chi1} * {RAD2DEG}"
        else:
            raise ValueError(f"Unsupported type for rot_chi1: {type(rot_chi1)}")
        
        if type(rot_chi2) is float:
            rot_chi2    = np.rad2deg(COORD_CHI2_FACTOR * rot_chi2)
        elif type(rot_chi2) is str:
            rot_chi2    = f"{COORD_CHI2_FACTOR} * {rot_chi2} * {RAD2DEG}"
        else:
            raise ValueError(f"Unsupported type for rot_chi2: {type(rot_chi2)}")
        
        if type(rot_chi3) is float:
            rot_chi3    = np.rad2deg(COORD_CHI3_FACTOR * rot_chi3)
        elif type(rot_chi3) is str:
            rot_chi3    = f"{COORD_CHI3_FACTOR} * {rot_chi3} * {RAD2DEG}"
        else:
            raise ValueError(f"Unsupported type for rot_chi3: {type(rot_chi3)}")

        ########################################
        # Compound Coordinate Transformation Element
        ########################################
        if n_transforms == 0:
            # In this case, it is some transform, but we don't know what, so guess this
            environment.new(
                name    = ele_name,
                parent  = xt.XYShift)
            print(f'Warning! Coordinate transformation {ele_name} has no transformations defined, installing as XYShift')
            continue
        elif n_transforms == 1:
            if offset_x != 0:
                environment.new(
                    name    = ele_name,
                    parent  = xt.XYShift,
                    dx      = offset_x)
            if offset_y != 0:
                environment.new(
                    name    = ele_name,
                    parent  = xt.XYShift,
                    dy      = offset_y)
            if rot_chi1 != 0:
                environment.new(
                    name    = ele_name,
                    parent  = xt.YRotation,
                    angle   = rot_chi1)
            if rot_chi2 != 0:
                environment.new(
                    name    = ele_name,
                    parent  = xt.XRotation,
                    angle   = rot_chi2)
            if rot_chi3 != 0:
                environment.new(
                    name    = ele_name,
                    parent  = xt.SRotation,
                    angle   = rot_chi3)
        elif n_transforms == 2 and offset_x != 0 and offset_y != 0:
            environment.new(
                name    = ele_name,
                parent  = xt.XYShift,
                dx      = offset_x,
                dy      = offset_y)
        else:
            compound_coord_transform_components = []
            # Order from testing and agrees with the SAD manual online

            # Transverse Shifts First
            if offset_x != 0 or offset_y != 0:
                environment.new(
                    name    = f'{ele_name}_dxy',
                    parent  = xt.XYShift,
                    dx      = offset_x,
                    dy      = offset_y)
                compound_coord_transform_components.append(f'{ele_name}_dxy')
            # YRotation Second
            if rot_chi1 != 0:
                environment.new(
                    name    = f'{ele_name}_chi1',
                    parent  = xt.YRotation,
                    angle   = rot_chi1)
                compound_coord_transform_components.append(f'{ele_name}_chi1')
            # XRotation Third
            if rot_chi2 != 0:
                environment.new(
                    name    = f'{ele_name}_chi2',
                    parent  = xt.XRotation,
                    angle   = rot_chi2)
                compound_coord_transform_components.append(f'{ele_name}_chi2')
            # SRotation Fourth
            if rot_chi3 != 0:
                environment.new(
                    name    = f'{ele_name}_chi3',
                    parent  = xt.SRotation,
                    angle   = rot_chi3)
                compound_coord_transform_components.append(f'{ele_name}_chi3')

            environment.new_line(
                name        = ele_name,
                components  = compound_coord_transform_components)
            continue


"""
(Unofficial) SAD to XSuite Converter

User Specified Element Exclusion
"""

################################################################################
# Required Packages
################################################################################
from .._globals import print_section_heading

################################################################################
# Exclude particular elements
################################################################################
def exclude_elements(
        parsed_lattice_data:    dict,
        excluded_elements:      list[str] | None    = None,
        verbose:                bool                = False) -> dict:
    
    ########################################
    # Check if there are excluded elements
    ########################################
    if verbose:
        print_section_heading("Checking for Excluded Elements", mode = 'subsection')
    if excluded_elements is None or len(excluded_elements) == 0:
        if verbose:
            print("No excluded elements found. Skipping exclusion.")
        return parsed_lattice_data
    
    ########################################
    # When we exclude elements, need to exclude the reverse also
    ########################################
    excluded_elements   += [
        elem[1:] for elem in excluded_elements if elem.startswith('-')]
    excluded_elements   += [
        "-" + elem for elem in excluded_elements if not elem.startswith('-')]

    ########################################
    # Get the required data
    ########################################
    parsed_elements     = parsed_lattice_data['elements']
    parsed_lines        = parsed_lattice_data['lines']

    ########################################
    # Delete the excluded elements from the elements dictionary
    ########################################
    for category, elems_dict in parsed_elements.items():
        # iterate over a snapshot of the keys
        for element in list(elems_dict.keys()):
            if element in excluded_elements:
                del elems_dict[element]
                if verbose:
                    print(f"Element {element} excluded from conversion")

    ########################################
    # Delete the excluded elements from the lines dictionary
    ########################################
    for line, components in parsed_lines.items():
        parsed_lines[line] = [comp for comp in components if comp not in excluded_elements]
    
    return parsed_lattice_data



"""
(Unofficial) SAD to XSuite Converter

Output Writer: Bends
"""

################################################################################
# Import Packages
################################################################################
from ._000_helpers import *

################################################################################
# Lattice File
################################################################################
def create_bend_lattice_file_information(line, line_table):

    ########################################
    # Get information
    ########################################
    hbends, vbends, sbends, unique_bend_variables, bend_name_dict = extract_bend_information(line, line_table)
    
    hbend_lengths       = np.array(sorted(hbends.keys()))
    hbend_names         = generate_magnet_for_replication_names(hbends, "hbend")
    vbend_lengths       = np.array(sorted(vbends.keys()))
    vbend_names         = generate_magnet_for_replication_names(vbends, "vbend")
    sbend_lengths       = np.array(sorted(sbends.keys()))
    sbend_names         = generate_magnet_for_replication_names(sbends, "sbend")

    ########################################
    # Ensure there are bends in the line
    ########################################
    if len(hbend_names) == 0 and len(vbend_names) == 0 and len(sbend_names) == 0:
        return ""

    ########################################
    # Create Output string
    ########################################
    output_string   = f"""
############################################################
# Bends
############################################################
"""

    ########################################
    # Create base elements
    ########################################
    output_string += f"""
########################################
# Base Elements
########################################"""

    for hbend_name, hbend_length in zip(hbend_names, hbend_lengths):
        output_string += f"""
env.new(name = '{hbend_name}', parent = xt.Bend, length = {hbend_length})"""

    for vbend_name, vbend_length in zip(vbend_names, vbend_lengths):
        output_string += f"""
env.new(name = '{vbend_name}', parent = xt.Bend, length = {vbend_length}, rot_s_rad = +np.pi/2)"""

    for sbend_name, sbend_length in zip(sbend_names, sbend_lengths):
        output_string += f"""
env.new(name = '{sbend_name}', parent = xt.Bend, length = {sbend_length})"""

    output_string += "\n"

    ########################################
    # Clone Elements
    ########################################
    output_string += f"""
########################################
# Cloned Elements
########################################"""

    for hbend, hbend_length in zip(hbend_names, hbend_lengths):
        for replica_name in hbends[hbend_length]:
            replica_variable    = bend_name_dict[replica_name]

            # If simple try to make it more compact
            if check_is_simple_bend_corr(line, replica_name):
                bend_generation = f"""
env.new(name = '{replica_name}', parent = '{hbend}', k0 = 'k0_{replica_variable}', h = {line[replica_name].h})"""

            # Otherwise do the full version
            else:
                bend_generation = f"""
env.new(
    name                    = '{replica_name}',
    parent                  = '{hbend}',
    k0                      = 'k0_{replica_variable}',
    h                       = {line[replica_name].h}"""
            # Append edge entry angles
                if line[replica_name].edge_entry_angle != 0:
                    bend_generation += f""",
    edge_entry_angle        = {line[replica_name].edge_entry_angle}"""
                if line[replica_name].edge_exit_angle != 0:
                    bend_generation += f""",
    edge_exit_angle         = {line[replica_name].edge_exit_angle}"""
                if line[replica_name].edge_entry_angle_fdown != 0:
                    bend_generation += f""",
    edge_entry_angle_fdown  = {line[replica_name].edge_entry_angle_fdown}"""
                if line[replica_name].edge_exit_angle_fdown != 0:
                    bend_generation += f""",
    edge_exit_angle_fdown   = {line[replica_name].edge_exit_angle_fdown}"""
                # Append shifts if they exist
                if line[replica_name].shift_x != 0:
                    bend_generation += f""",
    shift_x                 = '{line[replica_name].shift_x}'"""
                if line[replica_name].shift_y != 0:
                    bend_generation += f""",
    shift_y                 = '{line[replica_name].shift_y}'"""
                # Append the missing parenthesis
                bend_generation += """)"""
            
            # Write to the file
            output_string += bend_generation


    for vbend, vbend_length in zip(vbend_names, vbend_lengths):
        for replica_name in vbends[vbend_length]:
            replica_variable    = bend_name_dict[replica_name]

            # If simple try to make it more compact
            if check_is_simple_bend_corr(line, replica_name):
                bend_generation = f"""
env.new(name = '{replica_name}', parent = '{vbend}', k0 = 'k0_{replica_variable}', h = {line[replica_name].h})"""

            # Otherwise do the full version
            else:
                bend_generation = f"""
env.new(
    name                    = '{replica_name}',
    parent                  = '{vbend}',
    k0                      = 'k0_{replica_variable}',
    h                       = {line[replica_name].h}"""
            # Append edge entry angles
                if line[replica_name].edge_entry_angle != 0:
                    bend_generation += f""",
    edge_entry_angle        = {line[replica_name].edge_entry_angle}"""
                if line[replica_name].edge_exit_angle != 0:
                    bend_generation += f""",
    edge_exit_angle         = {line[replica_name].edge_exit_angle}"""
                if line[replica_name].edge_entry_angle_fdown != 0:
                    bend_generation += f""",
    edge_entry_angle_fdown  = {line[replica_name].edge_entry_angle_fdown}"""
                if line[replica_name].edge_exit_angle_fdown != 0:
                    bend_generation += f""",
    edge_exit_angle_fdown   = {line[replica_name].edge_exit_angle_fdown}"""
                # Append shifts if they exist
                if line[replica_name].shift_x != 0:
                    bend_generation += f""",
    shift_x                 = '{line[replica_name].shift_x}'"""
                if line[replica_name].shift_y != 0:
                    bend_generation += f""",
    shift_y                 = '{line[replica_name].shift_y}'"""
                # Append the missing parenthesis
                bend_generation += """)"""
            
            # Write to the file
            output_string += bend_generation

    for sbend, sbend_length in zip(sbend_names, sbend_lengths):
        for replica_name in sbends[sbend_length]:
            replica_variable    = bend_name_dict[replica_name]

            # If simple try to make it more compact
            if check_is_simple_bend_corr(line, replica_name):
                bend_generation = f"""
env.new(name = '{replica_name}', parent = '{sbend}', k0 = 'k0_{replica_variable}', h = {line[replica_name].h}, rot_s_rad = '{line[replica_name].rot_s_rad}')"""

            # Otherwise do the full version
            else:
                bend_generation = f"""
env.new(
    name                    = '{replica_name}',
    parent                  = '{sbend}',
    k0                      = 'k0_{replica_variable}',
    h                       = {line[replica_name].h}"""
            # Append edge entry angles
                if line[replica_name].edge_entry_angle != 0:
                    bend_generation += f""",
    edge_entry_angle        = {line[replica_name].edge_entry_angle}"""
                if line[replica_name].edge_exit_angle != 0:
                    bend_generation += f""",
    edge_exit_angle         = {line[replica_name].edge_exit_angle}"""
                if line[replica_name].edge_entry_angle_fdown != 0:
                    bend_generation += f""",
    edge_entry_angle_fdown  = {line[replica_name].edge_entry_angle_fdown}"""
                if line[replica_name].edge_exit_angle_fdown != 0:
                    bend_generation += f""",
    edge_exit_angle_fdown   = {line[replica_name].edge_exit_angle_fdown}"""
                # Append shifts if they exist
                if line[replica_name].shift_x != 0:
                    bend_generation += f""",
    shift_x                 = '{line[replica_name].shift_x}'"""
                if line[replica_name].shift_y != 0:
                    bend_generation += f""",
    shift_y                 = '{line[replica_name].shift_y}'"""
            # In the case of a skew corrector, we need to add a rotation
                bend_generation += f""",
    rot_s_rad               = '{line[replica_name].rot_s_rad}'"""
                # Append the missing parenthesis
                bend_generation += """)"""
            
            # Write to the file
            output_string += bend_generation

    ########################################
    # Return
    ########################################
    output_string += "\n"
    return output_string


################################################################################
# Optics File
################################################################################
def create_bend_optics_file_information(line, line_table):

    ########################################
    # Get information
    ########################################
    hbends, vbends, sbends, unique_bend_variables, bend_name_dict = extract_bend_information(line, line_table)

    hbend_names         = generate_magnet_for_replication_names(hbends, "hbend")
    vbend_names         = generate_magnet_for_replication_names(vbends, "vbend")
    sbend_names         = generate_magnet_for_replication_names(sbends, "sbend")

    ########################################
    # Ensure there are bends in the line
    ########################################
    if len(hbend_names) == 0 and len(vbend_names) == 0 and len(sbend_names) == 0:
        return ""

    ########################################
    # Create Output string
    ########################################
    output_string   = f"""
    ############################################################
    # Bends
    ############################################################"""

    for bend_variable in unique_bend_variables:
        k0 = None

        k0  = line[bend_variable].k0
        if k0 == 0:
            k0 = None

        if k0 is not None:
            output_string += f"""
    {f'k0_{bend_variable}'}{' ' * (OUTPUT_STRING_SEP - len(f'k0_{bend_variable}') + 4)}{'= '}{k0:.12f},"""

    ########################################
    # Return
    ########################################
    output_string += "\n"
    return output_string


"""
(Unofficial) SAD to XSuite Converter

Output Writer: Correctors
"""

################################################################################
# Import Packages
################################################################################
from ._000_helpers import *

################################################################################
# Lattice File
################################################################################
def create_corrector_lattice_file_information(line, line_table):

    ########################################
    # Get information
    ########################################
    hcorrs, vcorrs, scorrs, unique_corr_variables, corr_name_dict = extract_corrector_information(line, line_table)
    
    hcorr_lengths       = np.array(sorted(hcorrs.keys()))
    hcorr_names         = generate_magnet_for_replication_names(hcorrs, "hcorr")
    vcorr_lengths       = np.array(sorted(vcorrs.keys()))
    vcorr_names         = generate_magnet_for_replication_names(vcorrs, "vcorr")
    scorr_lengths       = np.array(sorted(scorrs.keys()))
    scorr_names         = generate_magnet_for_replication_names(scorrs, "scorr")

    ########################################
    # Ensure there are correctors in the line
    ########################################
    if len(hcorr_names) == 0 and len(vcorr_names) == 0 and len(scorr_names) == 0:
        return ""

    ########################################
    # Create Output string
    ########################################
    output_string   = f"""
############################################################
# Correctors
############################################################
"""

    ########################################
    # Create base elements
    ########################################
    output_string += f"""
########################################
# Base Elements
########################################"""

    for hcorr_name, hcorr_length in zip(hcorr_names, hcorr_lengths):
        output_string += f"""
env.new(name = '{hcorr_name}', parent = xt.Bend, length = {hcorr_length})"""

    for vcorr_name, vcorr_length in zip(vcorr_names, vcorr_lengths):
        output_string += f"""
env.new(name = '{vcorr_name}', parent = xt.Bend, length = {vcorr_length}, rot_s_rad = +np.pi/2)"""

    for scorr_name, scorr_length in zip(scorr_names, scorr_lengths):
        output_string += f"""
env.new(name = '{scorr_name}', parent = xt.Bend, length = {scorr_length})"""

    output_string += "\n"

    ########################################
    # Clone Elements
    ########################################
    output_string += f"""
########################################
# Cloned Elements
########################################"""

    for hcorr, hcorr_length in zip(hcorr_names, hcorr_lengths):
        for replica_name in hcorrs[hcorr_length]:
            replica_variable    = corr_name_dict[replica_name]

            # If simple try to make it more compact
            if check_is_simple_bend_corr(line, replica_name):
                corr_generation = f"""
env.new(name = '{replica_name}', parent = '{hcorr}', k0 = 'k0_{replica_variable}')"""
            
            # Otherwise do the full version
            else:
                corr_generation = f"""
env.new(
    name                    = '{replica_name}',
    parent                  = '{hcorr}',
    k0                      = 'k0_{replica_variable}'"""
            # Append edge entry angles
                if line[replica_name].edge_entry_angle != 0:
                    corr_generation += f""",
    edge_entry_angle        = {line[replica_name].edge_entry_angle}"""
                if line[replica_name].edge_exit_angle != 0:
                    corr_generation += f""",
    edge_exit_angle         = {line[replica_name].edge_exit_angle}"""
                if line[replica_name].edge_entry_angle_fdown != 0:
                    corr_generation += f""",
    edge_entry_angle_fdown  = {line[replica_name].edge_entry_angle_fdown}"""
                if line[replica_name].edge_exit_angle_fdown != 0:
                    corr_generation += f""",
    edge_exit_angle_fdown   = {line[replica_name].edge_exit_angle_fdown}"""
                # Append shifts if they exist
                if line[replica_name].shift_x != 0:
                    corr_generation += f""",
    shift_x                 = '{line[replica_name].shift_x}'"""
                if line[replica_name].shift_y != 0:
                    corr_generation += f""",
    shift_y                 = '{line[replica_name].shift_y}'"""
            # Append the missing parenthesis
                corr_generation += """)"""
            
            # Write to the file
            output_string += corr_generation


    for vcorr, vcorr_length in zip(vcorr_names, vcorr_lengths):
        for replica_name in vcorrs[vcorr_length]:
            replica_variable    = corr_name_dict[replica_name]

            # If simple try to make it more compact
            if check_is_simple_bend_corr(line, replica_name):
                corr_generation = f"""
env.new(name = '{replica_name}', parent = '{vcorr}', k0 = 'k0_{replica_variable}')"""
            
            # Otherwise do the full version
            else:
                corr_generation = f"""
env.new(
    name                    = '{replica_name}',
    parent                  = '{vcorr}',
    k0                      = 'k0_{replica_variable}'"""
            # Append edge entry angles
                if line[replica_name].edge_entry_angle != 0:
                    corr_generation += f""",
    edge_entry_angle        = {line[replica_name].edge_entry_angle}"""
                if line[replica_name].edge_exit_angle != 0:
                    corr_generation += f""",
    edge_exit_angle         = {line[replica_name].edge_exit_angle}"""
                if line[replica_name].edge_entry_angle_fdown != 0:
                    corr_generation += f""",
    edge_entry_angle_fdown  = {line[replica_name].edge_entry_angle_fdown}"""
                if line[replica_name].edge_exit_angle_fdown != 0:
                    corr_generation += f""",
    edge_exit_angle_fdown   = {line[replica_name].edge_exit_angle_fdown}"""
                # Append shifts if they exist
                if line[replica_name].shift_x != 0:
                    corr_generation += f""",
    shift_x                 = '{line[replica_name].shift_x}'"""
                if line[replica_name].shift_y != 0:
                    corr_generation += f""",
    shift_y                 = '{line[replica_name].shift_y}'"""
            # Append the missing parenthesis
                corr_generation += """)"""
            
            # Write to the file
            output_string += corr_generation

    for scorr, scorr_length in zip(scorr_names, scorr_lengths):
        for replica_name in scorrs[scorr_length]:
            replica_variable    = corr_name_dict[replica_name]

            # If simple try to make it more compact
            if check_is_simple_bend_corr(line, replica_name):
                corr_generation = f"""
env.new(name = '{replica_name}', parent = '{scorr}', k0 = 'k0_{replica_variable}', rot_s_rad = '{line[replica_name].rot_s_rad}')"""
            
            # Otherwise do the full version
            else:
                corr_generation = f"""
env.new(
    name                    = '{replica_name}',
    parent                  = '{scorr}',
    k0                      = 'k0_{replica_variable}'"""
            # Append edge entry angles
                if line[replica_name].edge_entry_angle != 0:
                    corr_generation += f""",
    edge_entry_angle        = {line[replica_name].edge_entry_angle}"""
                if line[replica_name].edge_exit_angle != 0:
                    corr_generation += f""",
    edge_exit_angle         = {line[replica_name].edge_exit_angle}"""
                if line[replica_name].edge_entry_angle_fdown != 0:
                    corr_generation += f""",
    edge_entry_angle_fdown  = {line[replica_name].edge_entry_angle_fdown}"""
                if line[replica_name].edge_exit_angle_fdown != 0:
                    corr_generation += f""",
    edge_exit_angle_fdown   = {line[replica_name].edge_exit_angle_fdown}"""
                # Append shifts if they exist
                if line[replica_name].shift_x != 0:
                    corr_generation += f""",
    shift_x                 = '{line[replica_name].shift_x}'"""
                if line[replica_name].shift_y != 0:
                    corr_generation += f""",
    shift_y                 = '{line[replica_name].shift_y}'"""
            # In the case of a skew corrector, we need to add a rotation
                corr_generation += f""",
    rot_s_rad               = '{line[replica_name].rot_s_rad}'"""
            # Append the missing parenthesis
                corr_generation += """)"""
            
            # Write to the file
            output_string += corr_generation

    ########################################
    # Return
    ########################################
    output_string += "\n"
    return output_string


################################################################################
# Optics File
################################################################################
def create_corrector_optics_file_information(line, line_table):

    ########################################
    # Get information
    ########################################
    hcorrs, vcorrs, scorrs, unique_corr_variables, corr_name_dict = extract_corrector_information(line, line_table)

    hcorr_names         = generate_magnet_for_replication_names(hcorrs, "hcorr")
    vcorr_names         = generate_magnet_for_replication_names(vcorrs, "vcorr")
    scorr_names         = generate_magnet_for_replication_names(scorrs, "scorr")

    ########################################
    # Ensure there are correctors in the line
    ########################################
    if len(hcorr_names) == 0 and len(vcorr_names) == 0 and len(scorr_names) == 0:
        return ""

    ########################################
    # Create Output string
    ########################################
    # TODO: This still gives an empty section is they are all set to 0
    output_string   = f"""
    ############################################################
    # Correctors
    ############################################################"""

    for corr_variable in unique_corr_variables:
        k0 = None

        k0  = line[corr_variable].k0
        if k0 == 0:
            k0 = None

        if k0 is not None:
            output_string += f"""
    {f'k0_{corr_variable}'}{' ' * (OUTPUT_STRING_SEP - len(f'k0_{corr_variable}') + 4)}{'= '}{k0:.12f},"""

    ########################################
    # Return
    ########################################
    output_string += "\n"
    return output_string


"""
(Unofficial) SAD to XSuite Converter
"""

################################################################################
# Required Packages
################################################################################
import os
import sad2xs as s2x
import xtrack as xt
import numpy as np
import textwrap
import matplotlib.pyplot as plt
from tqdm import tqdm

from _sad_helpers import sad_twiss, rebuild_sad_lattice
from _test_config import *

################################################################################
# N.B.
################################################################################
# REFERENCE SHIFTS INSIDE THE SOLENOID REGION DON'T WORK
# INSTEAD, WE USE MULTIPOLES TO SHIFT THE BEAM

################################################################################
# Reference PyTest Function
################################################################################
def reference_sol_test(
        test_name:                  str,
        sad_elements_line_string:   str,
        sad_eval_marker:            str         = "END",
        test_values:                np.ndarray  = TEST_VALUES,
        static_val:                 float       = STATIC_OFFSET,
        plot:                       bool        = True):
    """
    Reference Test the conversion of a SAD COORD element with DX and DY to XSuite.
    """

    ############################################################################
    # Scan
    ############################################################################
    S_SAD       = np.zeros_like(test_values)
    S_XS        = np.zeros_like(test_values)
    X_SAD       = np.zeros_like(test_values)
    X_XS        = np.zeros_like(test_values)
    Y_SAD       = np.zeros_like(test_values)
    Y_XS        = np.zeros_like(test_values)
    PX_SAD      = np.zeros_like(test_values)
    PX_XS       = np.zeros_like(test_values)
    PY_SAD      = np.zeros_like(test_values)
    PY_XS       = np.zeros_like(test_values)

    for iteration, test_val in enumerate(tqdm(test_values)):

        ########################################################################
        # Write Test Lattice
        ########################################################################
        with open("test_lattice.sad", "w") as f:
            f.write(textwrap.dedent(f"""\
            MOMENTUM    = 1.0 GEV;
            TEST_VAL    = {test_val};
            STATIC_VAL  = {static_val};
            """))

        with open("test_lattice.sad", "a") as f:
            f.write(sad_elements_line_string)

        ########################################################################
        # Twiss SAD Lattice
        ########################################################################
        tw_sad  = sad_twiss(
            lattice_filename    = 'test_lattice.sad',
            line_name           = 'TEST_LINE')

        ########################################################################
        # Rebuild SAD Lattice
        ########################################################################
        rebuild_sad_lattice(
            lattice_filename    = 'test_lattice.sad',
            line_name           = 'TEST_LINE')

        ########################################################################
        # Convert Lattice
        ########################################################################
        line    = s2x.convert_sad_to_xsuite(
            sad_lattice_path    = 'test_lattice.sad',
            output_directory    = None,
            _verbose            = False,
            _test_mode          = True)
        
        ########################################
        # Twiss XSuite Lattice
        ########################################
        tw_xs   = line.twiss4d(
            _continue_if_lost   = True,
            start               = xt.START,
            end                 = xt.END,
            betx                = 1,
            bety                = 1)

        ########################################################################
        # Diagnostics: Outputs if the test fails
        ########################################################################
        print("SAD")
        tw_sad.cols["s x px y py"].show()
        print("Xsuite")
        tw_xs.cols["s x px y py"].show()

        ########################################################################
        # Save data for comparison
        ########################################################################
        xs_eval_marker      = sad_eval_marker.lower()

        S_SAD[iteration]    = tw_sad["s", sad_eval_marker]
        S_XS[iteration]     = tw_xs["s", xs_eval_marker]
        X_SAD[iteration]    = tw_sad["x", sad_eval_marker]
        X_XS[iteration]     = tw_xs["x", xs_eval_marker]
        Y_SAD[iteration]    = tw_sad["y", sad_eval_marker]
        Y_XS[iteration]     = tw_xs["y", xs_eval_marker]
        PX_SAD[iteration]   = tw_sad["px", sad_eval_marker]
        PX_XS[iteration]    = tw_xs["px", xs_eval_marker]
        PY_SAD[iteration]   = tw_sad["py", sad_eval_marker]
        PY_XS[iteration]    = tw_xs["py", xs_eval_marker]

        ########################################################################
        # Delete test lattice
        ########################################################################
        os.remove("test_lattice.sad")

    ############################################################################
    # Plot
    ############################################################################
    if plot:

        symlog_threshold    = np.nanmin(
            np.where(TEST_VALUES > 0, TEST_VALUES, np.nan))

        if not np.all(np.isclose(S_SAD, S_XS, rtol = DELTA_S_RTOL, atol = DELTA_S_ATOL)):
            fig, axs    = plt.subplots(
                nrows       = 2,
                ncols       = 1,
                figsize     = (12, 6),
            sharex      = True,
            gridspec_kw = {'height_ratios': [3, 1]})

            axs[0].plot(TEST_VALUES, S_SAD, label = "SAD", color = "r")
            axs[0].plot(TEST_VALUES, S_XS, label = "XS", color = "b", linestyle='--')
            axs[1].plot(TEST_VALUES, S_XS - S_SAD)

            axs[1].set_xlabel("Drift Length [m]")
            axs[0].set_ylabel('s [m]')
            axs[1].set_ylabel('∆s [m]')

            for ax in axs:
                ax.set_xscale('symlog', linthresh = symlog_threshold)
                ax.set_yscale('symlog', linthresh = symlog_threshold)
                ax.grid()
            axs[0].legend()
            plt.subplots_adjust(hspace = 0)
            plt.savefig(f"outputs/{test_name}_s.png", dpi = 300, bbox_inches = 'tight')

        if not (
                np.all(np.isclose(X_SAD, X_XS, rtol = DELTA_X_RTOL, atol = DELTA_X_ATOL)) and
                np.all(np.isclose(Y_SAD, Y_XS, rtol = DELTA_Y_RTOL, atol = DELTA_Y_ATOL))):
            fig, axs    = plt.subplots(
                nrows       = 2,
                ncols       = 2,
                figsize     = (12, 6),
                sharex      = True,
            gridspec_kw = {'height_ratios': [3, 1]})

            axs[0, 0].plot(TEST_VALUES, X_SAD, label = "SAD", color = "r")
            axs[0, 0].plot(TEST_VALUES, X_XS, label = "XS", color = "b", linestyle='--')
            axs[1, 0].plot(TEST_VALUES, X_XS - X_SAD)

            axs[0, 1].plot(TEST_VALUES, Y_SAD, label = "SAD", color = "r")
            axs[0, 1].plot(TEST_VALUES, Y_XS, label = "XS", color = "b", linestyle='--')
            axs[1, 1].plot(TEST_VALUES, Y_XS - Y_SAD)

            axs[1, 0].set_xlabel("Transform Parameter")
            axs[1, 1].set_xlabel("Transform Parameter")
            axs[0, 0].set_ylabel('x [m]')
            axs[1, 0].set_ylabel('∆x [m]')
            axs[0, 1].set_ylabel('y [m]')
            axs[1, 1].set_ylabel('∆y [m]')

            for ax in axs.flatten():
                ax.set_xscale('symlog', linthresh = symlog_threshold)
                ax.set_yscale('symlog', linthresh = symlog_threshold)
                ax.grid()
            axs[0, 0].legend()
            axs[0, 1].legend()
            plt.subplots_adjust(hspace = 0)
            plt.savefig(f"outputs/{test_name}_xy.png", dpi = 300, bbox_inches = 'tight')

        if not (
                np.all(np.isclose(PX_SAD, PX_XS, rtol = DELTA_X_RTOL, atol = DELTA_X_ATOL)) and
                np.all(np.isclose(PY_SAD, PY_XS, rtol = DELTA_Y_RTOL, atol = DELTA_Y_ATOL))):
            fig, axs    = plt.subplots(
                nrows       = 2,
                ncols       = 2,
                figsize     = (12, 6),
                sharex      = True,
                gridspec_kw = {'height_ratios': [3, 1]})

            axs[0, 0].plot(TEST_VALUES, PX_SAD, label = "SAD", color = "r")
            axs[0, 0].plot(TEST_VALUES, PX_XS, label = "XS", color = "b", linestyle='--')
            axs[1, 0].plot(TEST_VALUES, PX_XS - PX_SAD)

            axs[0, 1].plot(TEST_VALUES, PY_SAD, label = "SAD", color = "r")
            axs[0, 1].plot(TEST_VALUES, PY_XS, label = "XS", color = "b", linestyle='--')
            axs[1, 1].plot(TEST_VALUES, PY_XS - PY_SAD)

            axs[1, 0].set_xlabel("Transform Parameter")
            axs[1, 1].set_xlabel("Transform Parameter")
            axs[0, 0].set_ylabel('px [m]')
            axs[1, 0].set_ylabel('∆px [m]')
            axs[0, 1].set_ylabel('py [m]')
            axs[1, 1].set_ylabel('∆py [m]')

            for ax in axs.flatten():
                ax.set_xscale('symlog', linthresh = symlog_threshold)
                ax.set_yscale('symlog', linthresh = symlog_threshold)
                ax.grid()
            axs[0, 0].legend()
            axs[0, 1].legend()
            plt.subplots_adjust(hspace = 0)
            plt.savefig(f"outputs/{test_name}_pxpy.png", dpi = 300, bbox_inches = 'tight')

        plt.close("all")

    ############################################################################
    # Assertions
    ############################################################################
    assert np.all(
        np.isclose(S_SAD, S_XS, rtol = DELTA_S_RTOL, atol = DELTA_S_ATOL)), \
        "s values do not match between SAD and XSuite."
    assert np.all(
        np.isclose(X_SAD, X_XS, rtol = DELTA_X_RTOL, atol = DELTA_X_ATOL)), \
        "x values do not match between SAD and XSuite."
    assert np.all(
        np.isclose(Y_SAD, Y_XS, rtol = DELTA_Y_RTOL, atol = DELTA_Y_ATOL)), \
        "y values do not match between SAD and XSuite."
    assert np.all(
        np.isclose(PX_SAD, PX_XS, rtol = DELTA_PX_RTOL, atol = DELTA_PX_ATOL)), \
        "px values do not match between SAD and XSuite."
    assert np.all(
        np.isclose(PY_SAD, PY_XS, rtol = DELTA_PY_RTOL, atol = DELTA_PY_ATOL)), \
        "py values do not match between SAD and XSuite."

################################################################################
# DXDY
################################################################################

########################################
# Forward
########################################
def test_sol_off_out_dxdy():
    """
    Test the conversion of a SAD COORD element with DX and DY to XSuite.
    """
    reference_sol_test(
        test_name                  = "test_006_sol_off_out_dxdy",
        sad_elements_line_string   = textwrap.dedent(f"""\
            DRIFT       SOL_DRIFT   = (L = 1.00);

            MULT        TEST_MULT1  = (K0 = +0.01 * TEST_VAL SK0 = +0.01 * TEST_VAL)
                        TEST_MULT2  = (K0 = -0.01 * TEST_VAL SK0 = -0.01 * TEST_VAL);

            SOL         SOL_IN      = (BZ = 0 BOUND =1)
                        SOL_OUT     = (BZ = 0 BOUND =1 GEO = 1);

            MARK        START       = ()
                        END         = ()
                        SOL_START   = ()
                        SOL_END     = ();

            LINE        TEST_LINE   = (START
                SOL_IN SOL_START TEST_MULT1 SOL_DRIFT TEST_MULT2 SOL_END SOL_OUT END);
            """),
        sad_eval_marker            = "END",
        test_values                = TEST_VALUES,
        static_val                 = STATIC_OFFSET,
        plot                       = True)
    
########################################
# Reverse
########################################
def test_sol_off_out_dxdy_rev():
    """
    Test the conversion of a SAD COORD element with DX and DY to XSuite.
    """
    reference_sol_test(
        test_name                  = "test_006_sol_off_out_dxdy_rev",
        sad_elements_line_string   = textwrap.dedent(f"""\
            DRIFT       SOL_DRIFT   = (L = 1.00);

            MULT        TEST_MULT1  = (K0 = +0.01 * TEST_VAL SK0 = +0.01 * TEST_VAL)
                        TEST_MULT2  = (K0 = -0.01 * TEST_VAL SK0 = -0.01 * TEST_VAL);

            SOL         SOL_IN      = (BZ = 0 BOUND =1)
                        SOL_OUT     = (BZ = 0 BOUND =1 GEO = 1);

            MARK        START       = ()
                        END         = ()
                        SOL_START   = ()
                        SOL_END     = ();

            LINE        TEST_LINE   = (START
                SOL_IN SOL_START TEST_MULT1 SOL_DRIFT TEST_MULT2 SOL_END -SOL_OUT END);
            """),
        sad_eval_marker            = "END",
        test_values                = TEST_VALUES,
        static_val                 = STATIC_OFFSET,
        plot                       = True)

################################################################################
# DPX
################################################################################

########################################
# Forward
########################################
def test_sol_off_out_dpx():
    """
    Test the conversion of a SAD COORD element with DX and DY to XSuite.
    """
    reference_sol_test(
        test_name                  = "test_006_sol_off_out_dpx",
        sad_elements_line_string   = textwrap.dedent(f"""\
            DRIFT       SOL_DRIFT   = (L = 1.00);

            MULT        TEST_MULT  = (K0 = 0.01 * TEST_VAL);

            SOL         SOL_IN      = (BZ = 0 BOUND =1)
                        SOL_OUT     = (BZ = 0 BOUND =1 GEO = 1);

            MARK        START       = ()
                        END         = ()
                        SOL_START   = ()
                        SOL_END     = ();

            LINE        TEST_LINE   = (START
                SOL_IN SOL_START SOL_DRIFT TEST_MULT SOL_END SOL_OUT END);
            """),
        sad_eval_marker            = "END",
        test_values                = TEST_VALUES,
        static_val                 = STATIC_OFFSET,
        plot                       = True)

########################################
# Reverse In
########################################
def test_sol_off_out_dpx_rev_in():
    """
    Test the conversion of a SAD COORD element with DX and DY to XSuite.
    """
    reference_sol_test(
        test_name                  = "test_006_sol_off_out_dpx_rev_in",
        sad_elements_line_string   = textwrap.dedent(f"""\
            DRIFT       SOL_DRIFT   = (L = 1.00);

            MULT        TEST_MULT  = (K0 = 0.01 * TEST_VAL);

            SOL         SOL_IN      = (BZ = 0 BOUND =1)
                        SOL_OUT     = (BZ = 0 BOUND =1 GEO = 1);

            MARK        START       = ()
                        END         = ()
                        SOL_START   = ()
                        SOL_END     = ();

            LINE        TEST_LINE   = (START
                -SOL_IN SOL_START SOL_DRIFT TEST_MULT SOL_END SOL_OUT END);
            """),
        sad_eval_marker            = "END",
        test_values                = TEST_VALUES,
        static_val                 = STATIC_OFFSET,
        plot                       = True)
    
########################################
# Reverse Out
########################################
def test_sol_off_out_dpx_rev_out():
    """
    Test the conversion of a SAD COORD element with DX and DY to XSuite.
    """
    reference_sol_test(
        test_name                  = "test_006_sol_off_out_dpx_rev_out",
        sad_elements_line_string   = textwrap.dedent(f"""\
            DRIFT       SOL_DRIFT   = (L = 1.00);

            MULT        TEST_MULT  = (K0 = 0.01 * TEST_VAL);

            SOL         SOL_IN      = (BZ = 0 BOUND =1)
                        SOL_OUT     = (BZ = 0 BOUND =1 GEO = 1);

            MARK        START       = ()
                        END         = ()
                        SOL_START   = ()
                        SOL_END     = ();

            LINE        TEST_LINE   = (START
                SOL_IN SOL_START SOL_DRIFT TEST_MULT SOL_END -SOL_OUT END);
            """),
        sad_eval_marker            = "END",
        test_values                = TEST_VALUES,
        static_val                 = STATIC_OFFSET,
        plot                       = True)
    
########################################
# Both Reversed
########################################
def test_sol_off_out_dpx_rev_both():
    """
    Test the conversion of a SAD COORD element with DX and DY to XSuite.
    """
    reference_sol_test(
        test_name                  = "test_006_sol_off_out_dpx_rev_both",
        sad_elements_line_string   = textwrap.dedent(f"""\
            DRIFT       SOL_DRIFT   = (L = 1.00);

            MULT        TEST_MULT  = (K0 = 0.01 * TEST_VAL);

            SOL         SOL_IN      = (BZ = 0 BOUND =1)
                        SOL_OUT     = (BZ = 0 BOUND =1 GEO = 1);

            MARK        START       = ()
                        END         = ()
                        SOL_START   = ()
                        SOL_END     = ();

            LINE        TEST_LINE   = (START
                -SOL_IN SOL_START SOL_DRIFT TEST_MULT SOL_END -SOL_OUT END);
            """),
        sad_eval_marker            = "END",
        test_values                = TEST_VALUES,
        static_val                 = STATIC_OFFSET,
        plot                       = True)

################################################################################
# DPY
################################################################################

########################################
# Forward
########################################
def test_sol_off_out_dpy():
    """
    Test the conversion of a SAD COORD element with DX and DY to XSuite.
    """
    reference_sol_test(
        test_name                  = "test_006_sol_off_out_dpy",
        sad_elements_line_string   = textwrap.dedent(f"""\
            DRIFT       SOL_DRIFT   = (L = 1.00);

            MULT        TEST_MULT  = (SK0 = 0.01 * TEST_VAL);

            SOL         SOL_IN      = (BZ = 0 BOUND =1)
                        SOL_OUT     = (BZ = 0 BOUND =1 GEO = 1);

            MARK        START       = ()
                        END         = ()
                        SOL_START   = ()
                        SOL_END     = ();

            LINE        TEST_LINE   = (START
                SOL_IN SOL_START SOL_DRIFT TEST_MULT SOL_END SOL_OUT END);
            """),
        sad_eval_marker            = "END",
        test_values                = TEST_VALUES,
        static_val                 = STATIC_OFFSET,
        plot                       = True)

########################################
# Reverse In
########################################
def test_sol_off_out_dpy_rev_in():
    """
    Test the conversion of a SAD COORD element with DX and DY to XSuite.
    """
    reference_sol_test(
        test_name                  = "test_006_sol_off_out_dpy_rev_in",
        sad_elements_line_string   = textwrap.dedent(f"""\
            DRIFT       SOL_DRIFT   = (L = 1.00);

            MULT        TEST_MULT  = (SK0 = 0.01 * TEST_VAL);

            SOL         SOL_IN      = (BZ = 0 BOUND =1)
                        SOL_OUT     = (BZ = 0 BOUND =1 GEO = 1);

            MARK        START       = ()
                        END         = ()
                        SOL_START   = ()
                        SOL_END     = ();

            LINE        TEST_LINE   = (START
                -SOL_IN SOL_START SOL_DRIFT TEST_MULT SOL_END SOL_OUT END);
            """),
        sad_eval_marker            = "END",
        test_values                = TEST_VALUES,
        static_val                 = STATIC_OFFSET,
        plot                       = True)
    
########################################
# Reverse Out
########################################
def test_sol_off_out_dpy_rev_out():
    """
    Test the conversion of a SAD COORD element with DX and DY to XSuite.
    """
    reference_sol_test(
        test_name                  = "test_006_sol_off_out_dpy_rev_out",
        sad_elements_line_string   = textwrap.dedent(f"""\
            DRIFT       SOL_DRIFT   = (L = 1.00);

            MULT        TEST_MULT  = (SK0 = 0.01 * TEST_VAL);

            SOL         SOL_IN      = (BZ = 0 BOUND =1)
                        SOL_OUT     = (BZ = 0 BOUND =1 GEO = 1);

            MARK        START       = ()
                        END         = ()
                        SOL_START   = ()
                        SOL_END     = ();

            LINE        TEST_LINE   = (START
                SOL_IN SOL_START SOL_DRIFT TEST_MULT SOL_END -SOL_OUT END);
            """),
        sad_eval_marker            = "END",
        test_values                = TEST_VALUES,
        static_val                 = STATIC_OFFSET,
        plot                       = True)
    
########################################
# Both Reversed
########################################
def test_sol_off_out_dpy_rev_both():
    """
    Test the conversion of a SAD COORD element with DX and DY to XSuite.
    """
    reference_sol_test(
        test_name                  = "test_006_sol_off_out_dpy_rev_both",
        sad_elements_line_string   = textwrap.dedent(f"""\
            DRIFT       SOL_DRIFT   = (L = 1.00);

            MULT        TEST_MULT  = (SK0 = 0.01 * TEST_VAL);

            SOL         SOL_IN      = (BZ = 0 BOUND =1)
                        SOL_OUT     = (BZ = 0 BOUND =1 GEO = 1);

            MARK        START       = ()
                        END         = ()
                        SOL_START   = ()
                        SOL_END     = ();

            LINE        TEST_LINE   = (START
                -SOL_IN SOL_START SOL_DRIFT TEST_MULT SOL_END -SOL_OUT END);
            """),
        sad_eval_marker            = "END",
        test_values                = TEST_VALUES,
        static_val                 = STATIC_OFFSET,
        plot                       = True)

################################################################################
# DXDYDPX
################################################################################

########################################
# Forward
########################################
def test_sol_off_out_dxdydpx():
    """
    Test the conversion of a SAD COORD element with DX and DY to XSuite.
    """
    reference_sol_test(
        test_name                  = "test_006_sol_off_out_dxdydpx",
        sad_elements_line_string   = textwrap.dedent(f"""\
            DRIFT       SOL_DRIFT   = (L = 1.00);

            MULT        TEST_MULT1  = (K0 = +0.01 * TEST_VAL SK0 = +0.01 * TEST_VAL)
                        TEST_MULT2  = (K0 = -0.01 * TEST_VAL SK0 = -0.01 * TEST_VAL)
                        TEST_MULT3  = (K0 = +0.01 * TEST_VAL);

            SOL         SOL_IN      = (BZ = 0 BOUND =1)
                        SOL_OUT     = (BZ = 0 BOUND =1 GEO = 1);

            MARK        START       = ()
                        END         = ()
                        SOL_START   = ()
                        SOL_END     = ();

            LINE        TEST_LINE   = (START
                SOL_IN SOL_START TEST_MULT1 SOL_DRIFT TEST_MULT2 TEST_MULT3 SOL_END SOL_OUT END);
            """),
        sad_eval_marker            = "END",
        test_values                = TEST_VALUES,
        static_val                 = STATIC_OFFSET,
        plot                       = True)
    
########################################
# Reverse In
########################################
def test_sol_off_out_dxdydpx_rev_in():
    """
    Test the conversion of a SAD COORD element with DX and DY to XSuite.
    """
    reference_sol_test(
        test_name                  = "test_006_sol_off_out_dxdydpx_rev_in",
        sad_elements_line_string   = textwrap.dedent(f"""\
            DRIFT       SOL_DRIFT   = (L = 1.00);

            MULT        TEST_MULT1  = (K0 = +0.01 * TEST_VAL SK0 = +0.01 * TEST_VAL)
                        TEST_MULT2  = (K0 = -0.01 * TEST_VAL SK0 = -0.01 * TEST_VAL)
                        TEST_MULT3  = (K0 = +0.01 * TEST_VAL);

            SOL         SOL_IN      = (BZ = 0 BOUND =1)
                        SOL_OUT     = (BZ = 0 BOUND =1 GEO = 1);

            MARK        START       = ()
                        END         = ()
                        SOL_START   = ()
                        SOL_END     = ();

            LINE        TEST_LINE   = (START
                -SOL_IN SOL_START TEST_MULT1 SOL_DRIFT TEST_MULT2 TEST_MULT3 SOL_END SOL_OUT END);
            """),
        sad_eval_marker            = "END",
        test_values                = TEST_VALUES,
        static_val                 = STATIC_OFFSET,
        plot                       = True)
    
########################################
# Reverse Out
########################################
def test_sol_off_out_dxdydpx_rev_out():
    """
    Test the conversion of a SAD COORD element with DX and DY to XSuite.
    """
    reference_sol_test(
        test_name                  = "test_006_sol_off_out_dxdydpx_rev_out",
        sad_elements_line_string   = textwrap.dedent(f"""\
            DRIFT       SOL_DRIFT   = (L = 1.00);

            MULT        TEST_MULT1  = (K0 = +0.01 * TEST_VAL SK0 = +0.01 * TEST_VAL)
                        TEST_MULT2  = (K0 = -0.01 * TEST_VAL SK0 = -0.01 * TEST_VAL)
                        TEST_MULT3  = (K0 = +0.01 * TEST_VAL);

            SOL         SOL_IN      = (BZ = 0 BOUND =1)
                        SOL_OUT     = (BZ = 0 BOUND =1 GEO = 1);

            MARK        START       = ()
                        END         = ()
                        SOL_START   = ()
                        SOL_END     = ();

            LINE        TEST_LINE   = (START
                SOL_IN SOL_START TEST_MULT1 SOL_DRIFT TEST_MULT2 TEST_MULT3 SOL_END -SOL_OUT END);
            """),
        sad_eval_marker            = "END",
        test_values                = TEST_VALUES,
        static_val                 = STATIC_OFFSET,
        plot                       = True)
    
########################################
# Both Reversed
########################################
def test_sol_off_out_dxdydpx_rev_both():
    """
    Test the conversion of a SAD COORD element with DX and DY to XSuite.
    """
    reference_sol_test(
        test_name                  = "test_006_sol_off_out_dxdydpx_rev_both",
        sad_elements_line_string   = textwrap.dedent(f"""\
            DRIFT       SOL_DRIFT   = (L = 1.00);

            MULT        TEST_MULT1  = (K0 = +0.01 * TEST_VAL SK0 = +0.01 * TEST_VAL)
                        TEST_MULT2  = (K0 = -0.01 * TEST_VAL SK0 = -0.01 * TEST_VAL)
                        TEST_MULT3  = (K0 = +0.01 * TEST_VAL);

            SOL         SOL_IN      = (BZ = 0 BOUND =1)
                        SOL_OUT     = (BZ = 0 BOUND =1 GEO = 1);

            MARK        START       = ()
                        END         = ()
                        SOL_START   = ()
                        SOL_END     = ();

            LINE        TEST_LINE   = (START
                -SOL_IN SOL_START TEST_MULT1 SOL_DRIFT TEST_MULT2 TEST_MULT3 SOL_END -SOL_OUT END);
            """),
        sad_eval_marker            = "END",
        test_values                = TEST_VALUES,
        static_val                 = STATIC_OFFSET,
        plot                       = True)

################################################################################
# DXDYDPY
################################################################################

########################################
# Forward
########################################
def test_sol_off_out_dxdydpy():
    """
    Test the conversion of a SAD COORD element with DX and DY to XSuite.
    """
    reference_sol_test(
        test_name                  = "test_006_sol_off_out_dxdydpy",
        sad_elements_line_string   = textwrap.dedent(f"""\
            DRIFT       SOL_DRIFT   = (L = 1.00);

            MULT        TEST_MULT1  = (K0 = +0.01 * TEST_VAL SK0 = +0.01 * TEST_VAL)
                        TEST_MULT2  = (K0 = -0.01 * TEST_VAL SK0 = -0.01 * TEST_VAL)
                        TEST_MULT3  = (SK0 = +0.01 * TEST_VAL);

            SOL         SOL_IN      = (BZ = 0 BOUND =1)
                        SOL_OUT     = (BZ = 0 BOUND =1 GEO = 1);

            MARK        START       = ()
                        END         = ()
                        SOL_START   = ()
                        SOL_END     = ();

            LINE        TEST_LINE   = (START
                SOL_IN SOL_START TEST_MULT1 SOL_DRIFT TEST_MULT2 TEST_MULT3 SOL_END SOL_OUT END);
            """),
        sad_eval_marker            = "END",
        test_values                = TEST_VALUES,
        static_val                 = STATIC_OFFSET,
        plot                       = True)
    
########################################
# Reverse In
########################################
def test_sol_off_out_dxdydpy_rev_in():
    """
    Test the conversion of a SAD COORD element with DX and DY to XSuite.
    """
    reference_sol_test(
        test_name                  = "test_006_sol_off_out_dxdydpy_rev_in",
        sad_elements_line_string   = textwrap.dedent(f"""\
            DRIFT       SOL_DRIFT   = (L = 1.00);

            MULT        TEST_MULT1  = (K0 = +0.01 * TEST_VAL SK0 = +0.01 * TEST_VAL)
                        TEST_MULT2  = (K0 = -0.01 * TEST_VAL SK0 = -0.01 * TEST_VAL)
                        TEST_MULT3  = (SK0 = +0.01 * TEST_VAL);

            SOL         SOL_IN      = (BZ = 0 BOUND =1)
                        SOL_OUT     = (BZ = 0 BOUND =1 GEO = 1);

            MARK        START       = ()
                        END         = ()
                        SOL_START   = ()
                        SOL_END     = ();

            LINE        TEST_LINE   = (START
                -SOL_IN SOL_START TEST_MULT1 SOL_DRIFT TEST_MULT2 TEST_MULT3 SOL_END SOL_OUT END);
            """),
        sad_eval_marker            = "END",
        test_values                = TEST_VALUES,
        static_val                 = STATIC_OFFSET,
        plot                       = True)
    
########################################
# Reverse Out
########################################
def test_sol_off_out_dxdydpy_rev_out():
    """
    Test the conversion of a SAD COORD element with DX and DY to XSuite.
    """
    reference_sol_test(
        test_name                  = "test_006_sol_off_out_dxdydpy_rev_out",
        sad_elements_line_string   = textwrap.dedent(f"""\
            DRIFT       SOL_DRIFT   = (L = 1.00);

            MULT        TEST_MULT1  = (K0 = +0.01 * TEST_VAL SK0 = +0.01 * TEST_VAL)
                        TEST_MULT2  = (K0 = -0.01 * TEST_VAL SK0 = -0.01 * TEST_VAL)
                        TEST_MULT3  = (SK0 = +0.01 * TEST_VAL);

            SOL         SOL_IN      = (BZ = 0 BOUND =1)
                        SOL_OUT     = (BZ = 0 BOUND =1 GEO = 1);

            MARK        START       = ()
                        END         = ()
                        SOL_START   = ()
                        SOL_END     = ();

            LINE        TEST_LINE   = (START
                SOL_IN SOL_START TEST_MULT1 SOL_DRIFT TEST_MULT2 TEST_MULT3 SOL_END -SOL_OUT END);
            """),
        sad_eval_marker            = "END",
        test_values                = TEST_VALUES,
        static_val                 = STATIC_OFFSET,
        plot                       = True)
    
########################################
# Both Reversed
########################################
def test_sol_off_out_dxdydpy_rev_both():
    """
    Test the conversion of a SAD COORD element with DX and DY to XSuite.
    """
    reference_sol_test(
        test_name                  = "test_006_sol_off_out_dxdydpy_rev_both",
        sad_elements_line_string   = textwrap.dedent(f"""\
            DRIFT       SOL_DRIFT   = (L = 1.00);

            MULT        TEST_MULT1  = (K0 = +0.01 * TEST_VAL SK0 = +0.01 * TEST_VAL)
                        TEST_MULT2  = (K0 = -0.01 * TEST_VAL SK0 = -0.01 * TEST_VAL)
                        TEST_MULT3  = (SK0 = +0.01 * TEST_VAL);

            SOL         SOL_IN      = (BZ = 0 BOUND =1)
                        SOL_OUT     = (BZ = 0 BOUND =1 GEO = 1);

            MARK        START       = ()
                        END         = ()
                        SOL_START   = ()
                        SOL_END     = ();

            LINE        TEST_LINE   = (START
                -SOL_IN SOL_START TEST_MULT1 SOL_DRIFT TEST_MULT2 TEST_MULT3 SOL_END -SOL_OUT END);
            """),
        sad_eval_marker            = "END",
        test_values                = TEST_VALUES,
        static_val                 = STATIC_OFFSET,
        plot                       = True)
    
################################################################################
# DXDYDPXDPY
################################################################################

########################################
# Forward
########################################
def test_sol_off_out_dxdydpxdpy():
    """
    Test the conversion of a SAD COORD element with DX and DY to XSuite.
    """
    reference_sol_test(
        test_name                  = "test_006_sol_off_out_dxdydpxdpy",
        sad_elements_line_string   = textwrap.dedent(f"""\
            DRIFT       SOL_DRIFT   = (L = 1.00);

            MULT        TEST_MULT  = (K0 = +0.01 * TEST_VAL SK0 = +0.01 * TEST_VAL);

            SOL         SOL_IN      = (BZ = 0 BOUND =1)
                        SOL_OUT     = (BZ = 0 BOUND =1 GEO = 1);

            MARK        START       = ()
                        END         = ()
                        SOL_START   = ()
                        SOL_END     = ();

            LINE        TEST_LINE   = (START
                SOL_IN SOL_START TEST_MULT1 SOL_DRIFT SOL_END SOL_OUT END);
            """),
        sad_eval_marker            = "END",
        test_values                = TEST_VALUES,
        static_val                 = STATIC_OFFSET,
        plot                       = True)

########################################
# Reverse In
########################################
def test_sol_off_out_dxdydpxdpy_rev_in():
    """
    Test the conversion of a SAD COORD element with DX and DY to XSuite.
    """
    reference_sol_test(
        test_name                  = "test_006_sol_off_out_dxdydpxdpy_rev_in",
        sad_elements_line_string   = textwrap.dedent(f"""\
            DRIFT       SOL_DRIFT   = (L = 1.00);

            MULT        TEST_MULT  = (K0 = +0.01 * TEST_VAL SK0 = +0.01 * TEST_VAL);

            SOL         SOL_IN      = (BZ = 0 BOUND =1)
                        SOL_OUT     = (BZ = 0 BOUND =1 GEO = 1);

            MARK        START       = ()
                        END         = ()
                        SOL_START   = ()
                        SOL_END     = ();

            LINE        TEST_LINE   = (START
                -SOL_IN SOL_START TEST_MULT1 SOL_DRIFT SOL_END SOL_OUT END);
            """),
        sad_eval_marker            = "END",
        test_values                = TEST_VALUES,
        static_val                 = STATIC_OFFSET,
        plot                       = True)

########################################
# Reverse Out
########################################
def test_sol_off_out_dxdydpxdpy_rev_out():
    """
    Test the conversion of a SAD COORD element with DX and DY to XSuite.
    """
    reference_sol_test(
        test_name                  = "test_006_sol_off_out_dxdydpxdpy_rev_out",
        sad_elements_line_string   = textwrap.dedent(f"""\
            DRIFT       SOL_DRIFT   = (L = 1.00);

            MULT        TEST_MULT  = (K0 = +0.01 * TEST_VAL SK0 = +0.01 * TEST_VAL);

            SOL         SOL_IN      = (BZ = 0 BOUND =1)
                        SOL_OUT     = (BZ = 0 BOUND =1 GEO = 1);

            MARK        START       = ()
                        END         = ()
                        SOL_START   = ()
                        SOL_END     = ();

            LINE        TEST_LINE   = (START
                SOL_IN SOL_START TEST_MULT1 SOL_DRIFT SOL_END -SOL_OUT END);
            """),
        sad_eval_marker            = "END",
        test_values                = TEST_VALUES,
        static_val                 = STATIC_OFFSET,
        plot                       = True)


########################################
# Both Reversed
########################################
def test_sol_off_out_dxdydpxdpy_rev_both():
    """
    Test the conversion of a SAD COORD element with DX and DY to XSuite.
    """
    reference_sol_test(
        test_name                  = "test_006_sol_off_out_dxdydpxdpy_rev_both",
        sad_elements_line_string   = textwrap.dedent(f"""\
            DRIFT       SOL_DRIFT   = (L = 1.00);

            MULT        TEST_MULT  = (K0 = +0.01 * TEST_VAL SK0 = +0.01 * TEST_VAL);

            SOL         SOL_IN      = (BZ = 0 BOUND =1)
                        SOL_OUT     = (BZ = 0 BOUND =1 GEO = 1);

            MARK        START       = ()
                        END         = ()
                        SOL_START   = ()
                        SOL_END     = ();

            LINE        TEST_LINE   = (START
                -SOL_IN SOL_START TEST_MULT1 SOL_DRIFT SOL_END -SOL_OUT END);
            """),
        sad_eval_marker            = "END",
        test_values                = TEST_VALUES,
        static_val                 = STATIC_OFFSET,
        plot                       = True)



"""
(Unofficial) SAD to XSuite Converter

Marker Converter
"""

################################################################################
# Required Packages
################################################################################
import numpy as np
import xtrack as xt

################################################################################
# Conversion Function
################################################################################
def convert_offset_markers(line, parsed_lattice_data):
    """
    Markers in SAD have an offset parameter that is not replicated in Xsuite
    """

    ########################################
    # Get the required data
    ########################################
    parsed_elements = parsed_lattice_data['elements']

    ########################################
    # Create output dictionary for the markers
    ########################################
    offset_marker_offsets   = {}

    ########################################
    # Get line table
    ########################################
    line.build_tracker()
    tt      = line.get_table(attr = True)
    line.discard_tracker()

    ########################################
    # Get the offsets for each marker
    ########################################
    print('Calculating offset marker positions')
    # Markers in Xsuite can come from mark, moni or beam-beam elements
    for marker_type in ['mark', 'moni', 'beambeam']:
        if marker_type in parsed_elements:
            for marker_name, marker in parsed_elements[marker_type].items():
                if 'offset' in marker:
                    offset_marker_offsets[marker_name] = marker['offset']

    ########################################
    # Get the names of the inserted markers in the line
    ########################################
    inserted_markers    = list(tt.rows[tt.element_type == 'Marker'].name)
    element_names       = list(tt.name)

    ########################################
    # Calculate intended marker locations
    ########################################
    offset_marker_locations = {}

    for marker in inserted_markers:

        base_marker     = marker.split('::')[0]

        if base_marker.startswith('-'):
            base_marker = base_marker[1:]
        
        ########################################
        # Only consider the offset markers
        ########################################
        if base_marker not in offset_marker_offsets:
            continue

        ########################################
        # Get the offset as a float
        ########################################
        offset  = offset_marker_offsets[base_marker]
        if isinstance(offset, str):
            # I think this is the source of the speed issue
            # but these can be strings with arithmetic expressions
            offset = eval(offset)

        ########################################
        # Case 1: Marker remains in the same element
        ########################################
        if 0 <= offset <= 1:
            marker_idx = element_names.index(marker)
            try:
                insert_at_ele   = element_names[marker_idx + 1]
                s_to_insert     = tt['s', insert_at_ele]
            except IndexError:  # Next element is the end of the line
                s_to_insert = tt.s[-1]
            except KeyError:    # Next element is a marker
                relative_idx = 1
                while True:
                    relative_idx += 1
                    try:
                        insert_at_ele   = element_names[marker_idx + relative_idx]
                        s_to_insert     = tt['s', insert_at_ele]
                        break
                    except KeyError:   # Next element is a marker
                        pass
                    except IndexError: # Next element is the end of the line
                        s_to_insert = tt.s[-1]
                        break

        ########################################
        # Case 2: Marker is offset to within another element
        ########################################
        else:
            # Get the index of the corresponding element
            relative_idx    = int(np.floor(offset))
            marker_idx      = element_names.index(marker)
            insert_at_ele   = element_names[marker_idx + relative_idx]

            # Get the length of the element to insert at
            insert_ele_length   = tt['length', insert_at_ele]

            # Add the fraction of element length
            s_to_insert     = tt['s', insert_at_ele] +\
                insert_ele_length * (offset % 1)

            ########################################
            # Exclusions!
            ########################################
            # TODO: Cannot slice magnets (currently unused)
            if isinstance(line[insert_at_ele], xt.Magnet):
                print("Cannot slice xt.Magnet element")
                print(f"Marker {base_marker} Ignored at {s_to_insert}")
                continue

            # TODO: Slicing solenoids causes issues
            if isinstance(line[insert_at_ele], xt.Solenoid):
                print("Slicing xt.Solenoid elements causes issues")
                print(f"Marker {base_marker} Ignored at {s_to_insert}")
                continue

        # Produce a dictionary of the s locations that markers are inserted at
        if base_marker in offset_marker_locations:
            offset_marker_locations[base_marker].append(s_to_insert)
        else:
            offset_marker_locations[base_marker] = [s_to_insert]

    ############################################################################
    # Remove the offset markers
    ############################################################################
    removed_markers = []
    for marker in inserted_markers:
        base_marker     = marker.split('::')[0]
        if base_marker.startswith('-'):
            base_marker = base_marker[1:]
        if base_marker not in offset_marker_offsets:
            continue

        if base_marker not in removed_markers:
            line.remove(base_marker)
            removed_markers.append(base_marker)

    ############################################################################
    # Return line
    ############################################################################
    return line, offset_marker_locations



"""
(Unofficial) SAD to XSuite Converter

Global and Deferred Expression Converter
"""

################################################################################
# Required Packages
################################################################################
import xtrack as xt

from .._globals import print_section_heading

################################################################################
# Parsing of strings and floats
################################################################################
def parse_expression(expression):
    """
    Try to convert s to float; if that fails, return s stripped
    """
    if type(expression) is float:
        return expression
    elif type(expression) is int:
        return float(expression)
    elif type(expression) is str:
        expression_stripped  = expression.strip()
        try:
            return float(expression_stripped)
        except ValueError:
            return expression_stripped
    else:
        raise TypeError(f"Unsupported type: {type(expression)}. Expected str, int, or float.")

################################################################################
# Convert Deferred Expressions
################################################################################
def convert_expressions(
        parsed_lattice_data:    dict,
        environment:            xt.Environment,
        verbose:                bool                = False) -> None:

    ########################################
    # Get the required data
    ########################################
    parsed_globals      = parsed_lattice_data['globals']
    parsed_expressions  = parsed_lattice_data['expressions']
    
    ########################################
    # Create global variables
    ########################################
    if verbose:
        print_section_heading("Converting Global Variable Expressions", mode = 'subsection')

    # Variables may depend on other variables, so have to parse them in order
    # Here, just try a few times to parse them
    converted_globals = []
    for i in range(10):
        for var_name, var_value in parsed_globals.items():
            
            if var_name in converted_globals:
                continue
            
            var_value   = parse_expression(var_value)
            try:
                environment[var_name] = var_value
                converted_globals.append(var_name)
            except KeyError:
                continue

    if len(converted_globals) != len(parsed_globals):
        raise ValueError("Not all global variables could be parsed. Check the input data.")

    ########################################
    # Create expressions
    ########################################
    if verbose:
        print_section_heading("Converting Deferred Expressions", mode = 'subsection')

    # Variables may depend on other variables, so have to parse them in order
    # Here, just try a few times to parse them
    converted_expressions = []
    for i in range(10):
        for var_name, var_value in parsed_expressions.items():

            if var_name in converted_expressions:
                continue
            
            var_value   = parse_expression(var_value)
            try:
                environment[var_name] = var_value
                converted_expressions.append(var_name)
            except KeyError:
                continue

    if len(converted_expressions) != len(parsed_expressions):
        raise ValueError("Not all expressions could be parsed. Check the input data.")


"""
(Unofficial) SAD to XSuite Converter
"""

################################################################################
# Import Packages
################################################################################
from datetime import date

from ..output_writer._002_bend import create_bend_optics_file_information
from ..output_writer._003_corr import create_corrector_optics_file_information
from ..output_writer._004_quad import create_quadrupole_optics_file_information
from ..output_writer._005_sext import create_sextupole_optics_file_information
from ..output_writer._006_oct import create_octupole_optics_file_information
from ..output_writer._009_cavity import create_cavity_optics_file_information
from ..output_writer._010_refshift import create_refshift_optics_file_information

today   = date.today()

################################################################################
# Write the optics file
################################################################################
def write_optics(
        line,
        output_filename,
        output_directory,
        output_header):
    """
    Write the outputs to the specified files.
    
    Parameters:
    line (xt.Line): The xtrack line object.
    output_filename (str): The base name for the output files.
    header (str): The header for the output files.
    """

    ########################################
    # Initialise the lattice file
    ########################################
    optics_file_string = f'''"""
{output_header}
================================================================================
Converted using the SAD2XS Converter
Authors:    J. Salvesen
Contact:    john.salvesen@cern.ch
================================================================================
Conversion Date: {today.strftime("%d/%m/%Y")}
"""

################################################################################
# Import Packages
################################################################################
import xtrack as xt

################################################################################
# Create Environment
################################################################################
env = xt.get_environment(verbose = True)

################################################################################
# Update Strengths
################################################################################
env.vars.update(default_to_zero = True,
'''

    ########################################
    # Add all the other sections
    ########################################
    line_table  = line.get_table(attr = True)

    optics_file_string  += create_bend_optics_file_information(line, line_table)
    optics_file_string  += create_corrector_optics_file_information(line, line_table)
    optics_file_string  += create_quadrupole_optics_file_information(line, line_table)
    optics_file_string  += create_sextupole_optics_file_information(line, line_table)
    optics_file_string  += create_octupole_optics_file_information(line, line_table)
    optics_file_string  += create_cavity_optics_file_information(line, line_table)
    optics_file_string  += create_refshift_optics_file_information(line, line_table)

    ########################################
    # Close the string
    ########################################
    optics_file_string  += ''')
'''

    ########################################
    # Write to file
    ########################################
    with open(f"{output_directory}/{output_filename}_import_optics.py", 'w') as f:
        f.write(optics_file_string)



"""
(Unofficial) SAD to XSuite Converter

User Specified Element Exclusion
"""

################################################################################
# Required Packages
################################################################################
from scipy.constants import c as clight

from .._globals import print_section_heading

################################################################################
# Exclude particular elements
################################################################################
def convert_harmonic_rf(
        line,
        parsed_lattice_data:    dict,
        verbose:                bool                = False) -> dict:
    
    ########################################
    # Check if the RF uses Harmonic Number
    ########################################
    if verbose:
        print_section_heading("Checking for RF using Harmonic Number", mode = 'subsection')

    has_cavities = 'cavi' in parsed_lattice_data["elements"]
    if not has_cavities:
        print("No cavities in line")
        return line

    has_harmonic_cavities   = any(
        isinstance(v, dict) and 'harm' in v
        for v in parsed_lattice_data["elements"]["cavi"].values())
    
    if not has_harmonic_cavities:
        print("No harmonic cavities in line")
        return line

    ########################################
    # Get the revolution frequency
    ########################################
    # Really we'd like the full one, but assume the circumference one
    circumference   = line.get_length()
    speed           = clight * line.particle_ref.beta0[0]

    f_rev           = speed / circumference

    ########################################
    # Go through the elements
    ########################################
    for cavity, properties in parsed_lattice_data["elements"]["cavi"].items():
        
        is_harmonic = isinstance(properties, dict) and 'harm' in properties
        
        if is_harmonic:
            harmonic_number    = properties['harm']
            frequency          = harmonic_number * f_rev

            if verbose:
                print(f"Converting cavity {cavity} with harmonic number {harmonic_number} to frequency {frequency:.3f} Hz")

            # Update in the line
            line[cavity].frequency = frequency

    return line



"""
(Unofficial) SAD to XSuite Converter

Reversal
"""

################################################################################
# Required Packages
################################################################################
import numpy as np

################################################################################
# Line Element Order Reversal
################################################################################
def reverse_line_element_order(line):
    """ Reverse the order of elements in a line and adjust their parameters
    accordingly to maintain the same physics but in the opposite direction.

    Parameters
    ----------
    line : xt.Line
        The original line to be reversed.

    Returns
    -------
    xt.Line
        A new line with elements in reverse order and adjusted parameters.
    """

    ########################################
    # Copy the line for changes
    ########################################
    env             = line.env
    env_elements    = list(set(env.elements.keys()))

    ########################################
    # Reverse Element Order
    ########################################
    line.mirror()

    ########################################
    # Get tables
    ########################################
    tt      = line.get_table(attr = True)
    tt_bend = tt.rows[
        (tt.element_type == 'Bend') | (tt.element_type == 'RBend')]
    tt_sol  = tt.rows[tt.element_type == 'Solenoid']
    tt_dxy  = tt.rows[tt.element_type == 'XYShift']

    ########################################
    # Get unique elements
    ########################################
    unique_bends    = list(set([name.split('::')[0] for name in tt_bend.name]))
    unique_sols     = list(set([name.split('::')[0] for name in tt_sol.name]))
    unique_dxys     = list(set([name.split('::')[0] for name in tt_dxy.name]))

    ########################################
    # Get only the non-reversed and handle reverse later
    ########################################
    # This only applies to the elements that can keep the minus sign
    unique_bends    = list(set(
        [name[1:] if name.startswith('-') else name for name in unique_bends]))
    unique_sols     = list(set(
        [name[1:] if name.startswith('-') else name for name in unique_sols]))
    unique_dxys     = list(set(
        [name[1:] if name.startswith('-') else name for name in unique_dxys]))

    ########################################
    # Bend Adjustments
    ########################################
    for bend in unique_bends:

        # Handling trying forward and reverse
        for bend in [bend, '-' + bend]:

            if bend not in env_elements:
                continue

            # Reverse entry/exit angles of bends
            entry_angle                 = env[bend].edge_entry_angle
            exit_angle                  = env[bend].edge_exit_angle
            env[bend].edge_entry_angle  = exit_angle
            env[bend].edge_exit_angle   = entry_angle

    ########################################
    # Solenoid Adjustments
    ########################################
    for sol in unique_sols:

        # Handling trying forward and reverse
        for sol in [sol, '-' + sol]:

            if sol not in env_elements:
                continue

            # Solenoid strength
            env[sol].ks *= -1

    ########################################
    # Reference Shifts
    ########################################
    for dxy in unique_dxys:

        # Handling trying forward and reverse
        for dxy in [dxy, '-' + dxy]:

            if dxy not in env_elements:
                continue
            env[dxy].dx *= -1
            env[dxy].dy *= -1

    return line

################################################################################
# Line Bend Direction Reversal
################################################################################
def reverse_line_bend_direction(line):
    """ Reverse the order of elements in a line and adjust their parameters
    accordingly to maintain the same physics but in the opposite direction.

    Parameters
    ----------
    line : xt.Line
        The original line to be reversed.

    Returns
    -------
    xt.Line
        A new line with elements in reverse order and adjusted parameters.
    """

    ########################################
    # Copy the line for changes
    ########################################
    env             = line.env
    env_elements    = list(set(env.elements.keys()))

    ########################################
    # Get tables
    ########################################
    tt      = line.get_table(attr = True)
    tt_bend = tt.rows[
        (tt.element_type == 'Bend') | (tt.element_type == 'RBend')]
    tt_quad = tt.rows[tt.element_type == 'Quadrupole']
    tt_sext = tt.rows[tt.element_type == 'Sextupole']
    tt_oct  = tt.rows[tt.element_type == 'Octupole']
    tt_mult = tt.rows[tt.element_type == 'Multipole']
    tt_sol  = tt.rows[tt.element_type == 'Solenoid']
    tt_dxy  = tt.rows[tt.element_type == 'XYShift']
    tt_chi1 = tt.rows[tt.element_type == 'YRotation']
    tt_chi2 = tt.rows[tt.element_type == 'XRotation']
    tt_chi3 = tt.rows[tt.element_type == 'SRotation']

    ########################################
    # Get unique elements
    ########################################
    unique_bends    = list(set([name.split('::')[0] for name in tt_bend.name]))
    unique_quads    = list(set([name.split('::')[0] for name in tt_quad.name]))
    unique_sexts    = list(set([name.split('::')[0] for name in tt_sext.name]))
    unique_octs     = list(set([name.split('::')[0] for name in tt_oct.name]))
    unique_mults    = list(set([name.split('::')[0] for name in tt_mult.name]))
    unique_sols     = list(set([name.split('::')[0] for name in tt_sol.name]))
    unique_dxys     = list(set([name.split('::')[0] for name in tt_dxy.name]))
    unique_chi1s    = list(set([name.split('::')[0] for name in tt_chi1.name]))
    unique_chi2s    = list(set([name.split('::')[0] for name in tt_chi2.name]))
    unique_chi3s    = list(set([name.split('::')[0] for name in tt_chi3.name]))

    ########################################
    # Get only the non-reversed and handle reverse later
    ########################################
    # This only applies to the elements that can keep the minus sign
    unique_bends    = list(set(
        [name[1:] if name.startswith('-') else name for name in unique_bends]))
    unique_sols     = list(set(
        [name[1:] if name.startswith('-') else name for name in unique_sols]))
    unique_dxys     = list(set(
        [name[1:] if name.startswith('-') else name for name in unique_dxys]))
    unique_chi1s    = list(set(
        [name[1:] if name.startswith('-') else name for name in unique_chi1s]))
    unique_chi2s    = list(set(
        [name[1:] if name.startswith('-') else name for name in unique_chi2s]))
    unique_chi3s    = list(set(
        [name[1:] if name.startswith('-') else name for name in unique_chi3s]))

    ########################################
    # Bend Adjustments
    ########################################
    for bend in unique_bends:

        # Handling trying forward and reverse
        for bend in [bend, '-' + bend]:

            if bend not in env_elements:
                continue
            
            env[bend].k0  *= -1
            env[bend].k1  *= +1
            env[bend].h   *= -1

            # Reverse entry/exit angles of bends
            env[bend].edge_entry_angle  *= -1
            env[bend].edge_exit_angle   *= -1

            # knl ksl Adjustments
            order   = env[bend]._order
            for even_order in np.arange(0, order + 1, 2):
                env[bend].knl[even_order] *= -1
                env[bend].ksl[even_order] *= +1
            for odd_order in np.arange(1, order + 1, 2):
                env[bend].knl[odd_order]  *= +1
                env[bend].ksl[odd_order]  *= -1

            # Offset adjustments
            env[bend].shift_x   *= -1
            env[bend].shift_y   *= +1
            env[bend].rot_s_rad *= -1

    ########################################
    # Quadrupole Adjustments
    ########################################
    for quad in unique_quads:

        env[quad].k1  *= +1
        env[quad].k1s *= -1

        # knl ksl Adjustments
        order   = env[quad]._order
        for even_order in np.arange(0, order + 1, 2):
            env[quad].knl[even_order] *= -1
            env[quad].ksl[even_order] *= +1
        for odd_order in np.arange(1, order + 1, 2):
            env[quad].knl[odd_order]  *= +1
            env[quad].ksl[odd_order]  *= -1

        # Offset adjustments
        env[quad].shift_x   *= -1
        env[quad].shift_y   *= +1
        env[quad].rot_s_rad *= -1

    ########################################
    # Sextupole Adjustments
    ########################################
    for sext in unique_sexts:

        env[sext].k2  *= -1
        env[sext].k2s *= +1

        # knl ksl Adjustments
        order   = env[sext]._order
        for even_order in np.arange(0, order + 1, 2):
            env[sext].knl[even_order] *= -1
            env[sext].ksl[even_order] *= +1
        for odd_order in np.arange(1, order + 1, 2):
            env[sext].knl[odd_order]  *= +1
            env[sext].ksl[odd_order]  *= -1

        # Offset adjustments
        env[sext].shift_x   *= -1
        env[sext].shift_y   *= +1
        env[sext].rot_s_rad *= -1

    ########################################
    # Octupole Adjustments
    ########################################
    for oct in unique_octs:

        env[oct].k3     *= +1
        env[oct].k3s    *= -1

        # knl ksl Adjustments
        order   = env[oct]._order
        for even_order in np.arange(0, order + 1, 2):
            env[oct].knl[even_order]  *= -1
            env[oct].ksl[even_order]  *= +1
        for odd_order in np.arange(1, order + 1, 2):
            env[oct].knl[odd_order]   *= +1
            env[oct].ksl[odd_order]   *= -1

        # Offset adjustments
        env[oct].shift_x    *= -1
        env[oct].shift_y    *= +1
        env[oct].rot_s_rad  *= -1

    ########################################
    # Multipole Adjustments
    ########################################
    for mult in unique_mults:

        # knl ksl Adjustments
        order   = env[mult]._order
        for even_order in np.arange(0, order + 1, 2):
            env[mult].knl[even_order] *= -1
            env[mult].ksl[even_order] *= +1
        for odd_order in np.arange(1, order + 1, 2):
            env[mult].knl[odd_order]  *= +1
            env[mult].ksl[odd_order]  *= -1

        # Offset adjustments
        env[mult].shift_x   *= -1
        env[mult].shift_y   *= +1
        env[mult].rot_s_rad *= -1

    ########################################
    # Solenoid Adjustments
    ########################################
    for sol in unique_sols:

        # Handling trying forward and reverse
        for sol in [sol, '-' + sol]:

            if sol not in env_elements:
                continue

            # Solenoid strength
            env[sol].ks *= -1

            # knl ksl Adjustments
            order   = env[sol]._order
            for even_order in np.arange(0, order + 1, 2):
                env[sol].knl[even_order]  *= -1
                env[sol].ksl[even_order]  *= +1
            for odd_order in np.arange(1, order + 1, 2):
                env[sol].knl[odd_order]   *= +1
                env[sol].ksl[odd_order]   *= -1

            # Offset adjustments
            env[sol].mult_shift_x   *= -1
            env[sol].mult_shift_y   *= +1
            env[sol].rot_s_rad      *= -1

    ########################################
    # Reference Shifts
    ########################################
    for dxy in unique_dxys:

        # Handling trying forward and reverse
        for dxy in [dxy, '-' + dxy]:

            if dxy not in env_elements:
                continue
            env[dxy].dx *= -1
            env[dxy].dy *= +1

    for chi1 in unique_chi1s:

        # Handling trying forward and reverse
        for chi1 in [chi1, '-' + chi1]:
            
            if chi1 not in env_elements:
                continue
            env[chi1].angle *= -1

    for chi2 in unique_chi2s:

        # Handling trying forward and reverse
        for chi2 in [chi2, '-' + chi2]:
            
            if chi2 not in env_elements:
                continue
            env[chi2].angle *= +1

    for chi3 in unique_chi3s:

        # Handling trying forward and reverse
        for chi3 in [chi3, '-' + chi3]:
            
            if chi3 not in env_elements:
                continue
            env[chi3].angle *= -1

    return line

################################################################################
# Charge Reversal
################################################################################
def reverse_line_charge(line):
    """ Reverse the order of elements in a line and adjust their parameters
    accordingly to maintain the same physics but in the opposite direction.

    Parameters
    ----------
    line : xt.Line
        The original line to be reversed.

    Returns
    -------
    xt.Line
        A new line with elements in reverse order and adjusted parameters.
    """

    ########################################
    # Flip the charge
    ########################################
    line.particle_ref.q0  *= -1

    return line


"""
Unofficial SAD to XSuite Lattice Converter
=============================================
Author(s):  John P T Salvesen
Email:      john.salvesen@cern.ch
Date:       30-09-2025
"""

################################################################################
# Required Packages
################################################################################
import xtrack as xt

from ._globals import print_section_heading, ASCII_LOGO, N_INTEGRATOR_KICKS

from .converter._001_parser import parse_sad_file
from .converter._002_element_exclusion import exclude_elements
from .converter._003_expression_converter import convert_expressions
from .converter._004_element_converter import convert_elements
from .converter._005_line_converter import convert_lines
from .converter._006_solenoid_converter import convert_solenoids, solenoid_reference_shift_corrections
from .converter._007_harmonic_rf import convert_harmonic_rf
from .converter._008_reversals import reverse_line_bend_direction, reverse_line_element_order, reverse_line_charge
from .converter._009_offset_markers import convert_offset_markers
from .converter._010_write_lattice import write_lattice
from .converter._011_write_optics import write_optics

################################################################################
# Overall Function
################################################################################
def convert_sad_to_xsuite(
        sad_lattice_path,
        output_directory,
        output_filename             = None,
        line_name                   = None,
        output_header               = "SAD to XSuite Lattice Conversion",
        excluded_elements           = None,
        user_multipole_replacements = None,
        reverse_element_order       = False,
        reverse_bend_direction      = False,
        reverse_charge              = False,
        _verbose                    = True,
        _test_mode                  = False):

    ############################################################################
    # Introduction Printout
    ############################################################################
    if _verbose:
        print(ASCII_LOGO)
        print(f"Processing SAD file: {sad_lattice_path}")

    ############################################################################
    # Parse Lattice
    ############################################################################
    if _verbose:
        print_section_heading("Parsing SAD File", mode = 'section')

    parsed_lattice_data = parse_sad_file(
        sad_lattice_path    = sad_lattice_path,
        ref_particle_mass0  = None,
        ref_particle_q0     = None,
        ref_particle_p0c    = None,
        verbose             = _verbose)

    ############################################################################
    # Remove Excluded elements
    ############################################################################
    if _verbose:
        print_section_heading("Removing Excluded Elements", mode = 'section')

    parsed_lattice_data = exclude_elements(
        parsed_lattice_data = parsed_lattice_data,
        excluded_elements   = excluded_elements,
        verbose             = _verbose)

    ############################################################################
    # Build Environment
    ############################################################################
    if _verbose:
        print_section_heading("Building Environment", mode = 'section')

    env = xt.Environment()

    ############################################################################
    # Convert Expressions
    ############################################################################
    if _verbose:
        print_section_heading("Converting Expressions", mode = 'section')

    convert_expressions(
        parsed_lattice_data = parsed_lattice_data,
        environment         = env,
        verbose             = _verbose)

    ########################################
    # Add reference particle from globals
    ########################################
    env.particle_ref    = xt.Particles(
        p0c     = parsed_lattice_data['globals']['momentum'],
        q0      = parsed_lattice_data['globals']['charge'],
        mass0   = parsed_lattice_data['globals']['mass'])

    ############################################################################
    # Convert Elements
    ############################################################################
    if _verbose:
        print_section_heading("Converting Elements", mode = 'section')

    convert_elements(
        parsed_lattice_data         = parsed_lattice_data,
        environment                 = env,
        user_multipole_replacements = user_multipole_replacements,
        verbose                     = _verbose)

    ############################################################################
    # Convert Lines
    ############################################################################
    if _verbose:
        print_section_heading("Converting Lines", mode = 'section')

    convert_lines(
        parsed_lattice_data = parsed_lattice_data,
        environment         = env)
    
    ########################################
    # Select the line
    ########################################
    if _verbose:
        print_section_heading("Selecting Line", mode = 'subsection')
    
    if line_name is not None:
        line = env.lines[line_name.lower()]
        if _verbose:
            print(f"Selected line: {line_name}")
    else:
        line_lengths    = {line: env.lines[line].get_length() for line in env.lines}
        
        # If several are the same length, check also number of elements (thin elements)
        if max(line_lengths.values()) != 0:
            longest_line    = max(line_lengths, key = lambda line: line_lengths[line])
        else:
            line_lengths    = {line: len(env.lines[line].element_names) for line in env.lines}
            longest_line    = max(line_lengths, key = lambda line: line_lengths[line])
        
        line            = env.lines[longest_line]

        if _verbose:
            print(f"Selected line: {longest_line}")

    ############################################################################
    # Solenoid Corrections
    ############################################################################
    if _verbose:
        print_section_heading("Performing Solenoid Corrections", mode = 'section')

    ########################################
    # Convert elements between solenoids
    ########################################
    if _verbose:
        print_section_heading("Converting Elements between Solenoids", mode = 'subsection')
    convert_solenoids(
        parsed_lattice_data = parsed_lattice_data,
        environment         = env,
        verbose             = _verbose)
    
    ########################################
    # Correct solenoid reference shifts
    ########################################
    if _verbose:
        print_section_heading("Correcting Solenoid Reference Shifts", mode = 'subsection')
    solenoid_reference_shift_corrections(
        line                    = line,
        parsed_lattice_data     = parsed_lattice_data,
        environment             = env,
        reverse_line            = reverse_element_order,
        verbose                 = _verbose)
    
    ############################################################################
    # Harmonic Cavity Correction
    ############################################################################
    if _verbose:
        print_section_heading("Converting Harmonic Cavities", mode = 'section')
    convert_harmonic_rf(
        line                = line,
        parsed_lattice_data = parsed_lattice_data,
        verbose             = _verbose)

    ################################################################################
    # Configure Modelling Mode
    ################################################################################
    if _verbose:
        print_section_heading("Configuring Modelling Mode", mode = 'section')

    ########################################
    # Use exact drifts
    ########################################
    # Needed for large angles inside drifts
    line.config.XTRACK_USE_EXACT_DRIFTS = True                  # type: ignore

    ########################################
    # Set bend model
    ########################################
    if _verbose:
        print_section_heading("Configuring Bend Model", mode = 'subsection')
    
    line.configure_bend_model(edge = 'full')

    ########################################
    # Set integrators
    ########################################
    if _verbose:
        print_section_heading("Configuring Integrators", mode = 'subsection')
    
    tt          = line.get_table()
    tt_drift    = tt.rows[tt.element_type == 'Drift']
    tt_bend     = tt.rows[tt.element_type == 'Bend']
    tt_quad     = tt.rows[tt.element_type == 'Quadrupole']
    tt_sext     = tt.rows[tt.element_type == 'Sextupole']
    tt_oct      = tt.rows[tt.element_type == 'Octupole']

    line.set(
        tt_drift,
        model               = 'exact')
    line.set(
        tt_bend,
        integrator          = 'uniform',
        num_multipole_kicks = N_INTEGRATOR_KICKS,
        model               = 'mat-kick-mat')
    line.set(
        tt_quad,
        integrator          = 'uniform',
        num_multipole_kicks = N_INTEGRATOR_KICKS,
        model               = 'mat-kick-mat')
    line.set(
        tt_sext,
        integrator          = 'yoshida4',
        num_multipole_kicks = N_INTEGRATOR_KICKS)
    line.set(
        tt_oct,
        integrator          = 'yoshida4',
        num_multipole_kicks = N_INTEGRATOR_KICKS)

    ############################################################################
    # Line reversals
    ############################################################################
    if reverse_element_order:
        if _verbose:
            print_section_heading("Reversing Element order of Line", mode = 'section')
        line = reverse_line_element_order(line)

    if reverse_bend_direction:
        if _verbose:
            print_section_heading("Reversing Bend Directions of Line", mode = 'section')
        line = reverse_line_bend_direction(line)

    if reverse_charge:
        if _verbose:
            print_section_heading("Reversing Charge of Line", mode = 'section')
        line = reverse_line_charge(line)

    ############################################################################
    # Handle Offset Markers
    ############################################################################
    if _verbose:
        print_section_heading("Converting Offset Markers", mode = 'subsection')

    line, offset_marker_locations   = convert_offset_markers(
        line                = line,
        parsed_lattice_data = parsed_lattice_data)

    ############################################################################
    # Breakpoint for testing
    ############################################################################
    if _test_mode:
        return line

    ############################################################################
    # Output files
    ############################################################################

    ########################################
    # Filename
    ########################################
    if output_filename is None:
        output_filename = sad_lattice_path.split('/')[-1].replace('.sad', '')
    else:
        assert isinstance(output_filename, str), "output_filename must be a string"

    ########################################
    # Lattice
    ########################################
    if _verbose:
        print_section_heading("Generating Lattice File", mode = 'subsection')

    write_lattice(
        line                        = line,
        offset_marker_locations     = offset_marker_locations,
        output_filename             = output_filename,
        output_directory            = output_directory,
        output_header               = output_header)
    
    ########################################
    # Import optics
    ########################################
    if _verbose:
        print_section_heading("Generating Optics File", mode = 'subsection')

    write_optics(
        line                        = line,
        output_filename             = output_filename,
        output_directory            = output_directory,
        output_header               = output_header)

    ############################################################################
    # Delete and re-initialise
    ############################################################################

    ########################################
    # Delete messy import environment
    ########################################
    del env
    del line

    ########################################
    # Cleanly load from the generated files
    ########################################
    env     = xt.Environment()
    env.call(f"{output_directory}/{output_filename}.py")
    env.call(f"{output_directory}/{output_filename}_import_optics.py")
    line    = env.lines["line"]

    ############################################################################
    # Complete message
    ############################################################################
    if _verbose:
        print_section_heading("Conversion Complete", mode = 'section')

    ############################################################################
    # Return the line
    ############################################################################
    return line
